import re, json, httpx
from bs4 import BeautifulSoup
from typing import Callable

from ptlibs.ptprinthelper import get_colored_text, help_calc_column_width


class BlacklistParser:
    def __init__(self, ptdebug: Callable[..., None], use_json=None, verbose_mode=None):
        self.ptdebug = ptdebug
        self.use_json = use_json
        self.verbose_mode = verbose_mode
        self.result = None

    BLACKLIST_TIMEOUT = 15.0
    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"

    def _get_auth_keys(self):
        try:
            response = httpx.get(
                "https://mxtoolbox.com/api/v1/user",
                headers={
                    "Accept": "application/json, text/javascript, */*; q=0.01",
                    "User-Agent": self.USER_AGENT,
                    "Referer": "https://mxtoolbox.com/",
                },
                timeout=self.BLACKLIST_TIMEOUT,
            )
            if response.status_code != 200:
                raise ValueError(
                    f"mxtoolbox.com returned HTTP {response.status_code} "
                    f"(API may block automated access - try again later or check network)"
                )
            auth_key = response.json()
        except ValueError:
            raise
        except Exception as e:
            raise ValueError(f"Could not retrieve blacklist data: {type(e).__name__}: {e}") from e
        return auth_key["MxVisitorUid"], auth_key["TempAuthKey"]

    def _parse_response(self, response):
        soup = BeautifulSoup(response["HTML_Value"], "lxml")

        title = re.search(r"Checking([\s\S])*?blacklists\.", soup.text)[0]
        listed_info = re.search(r"Listed \d? times with \d? timeouts", soup.text)[0]
        result = {"title": title, "listed_info": listed_info, "table_result": []}

        table = soup.find("tbody")
        for table_a_href_tag in table.find_all("a"):
            table_a_href_tag.decompose()

        table_rows = table.find_all("tr")
        for table_row in table_rows:
            status = table_row.find("td", {"class": "table-column-Status"})
            blacklist_name = table_row.find("td", {"class": "table-column-Name"})
            reason = table_row.find("td", {"class": "tool-blacklist-reason"})
            ttl = table_row.find("td", {"class": "table-column-TTL"})
            # response_time = table_row.find('td', {'class': 'table-column-ResponseTime'})
            row_result = []
            for index, col in enumerate([status.text, blacklist_name.text, reason.text, ttl.text]):
                col.replace("\xa0", "")
                if col == "":
                    col = "-"
                row_result.append(col)
            row_result = [data.replace("\xa0", "") for data in row_result]
            # input(row_result)
            result["table_result"].append(row_result)

        return result

    def lookup(self, target):
        if self.verbose_mode:
            self.ptdebug("Retrieving auth keys ...", title=True)
        auth_visitor_id, auth_key = self._get_auth_keys()
        headers = {
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Tempauthorization": auth_key,
            "User-Agent": self.USER_AGENT,
            "Referer": "https://mxtoolbox.com/",
        }
        cookie = {"MxVisitorUID": auth_visitor_id}
        client = httpx.Client(http2=True, timeout=self.BLACKLIST_TIMEOUT)
        if self.verbose_mode:
            self.ptdebug("Retrieving server response ...", title=True)
        response = client.get(
            f"https://mxtoolbox.com/api/v1/Lookup?command=blacklist&argument={target}&resultIndex=1&disableRhsbl=true&format=2",
            headers=headers,
            cookies=cookie,
        ).json()
        if self.verbose_mode:
            self.ptdebug("Parsing response ... ", title=True, end="")
        if re.search("<h3>Private IP Address</h3>", response["HTML_Value"]):
            return "Cannot test Private IP Address"  # False
        result = self._parse_response(response)
        if self.verbose_mode:
            self.ptdebug("OK")

        self.result = result

        self._print_result(result)

    def get_response_json(self, result):
        if not result:
            raise ValueError("lookup method must be called first")
        return json.dumps(result)

    def _print_result(self, result):
        max_width = help_calc_column_width(result["table_result"])
        table_heading = f"STATUS{' '*(max_width[0]-6+5)}BLACKLIST{' '*(max_width[1]-9+2)}REASON{' '*(max_width[2]-6+7)}TTL{' '*(max_width[3]-3+2)}"
        self.ptdebug(get_colored_text("-" * (len(table_heading) + max_width[3] + 1), color="TITLE"))
        self.ptdebug(get_colored_text(table_heading, color="TITLE"))
        for table_row in result["table_result"]:
            table_row_before_colors = table_row[0]
            if table_row[0] == "OK":
                status = get_colored_text(table_row[0], color="NOTVULN")
            elif table_row[0] in ["LISTED", "TIMEOUT"]:
                status = get_colored_text(table_row[0], color="VULN")
            else:
                status = table_row[0]
            self.ptdebug(f"{status}", end=" " * (max_width[0] - len(table_row_before_colors) + 5))
            self.ptdebug(f"{table_row[1]}", end=" " * (max_width[1] - len(table_row[1]) + 2))
            self.ptdebug(f"{table_row[2]}", end=" " * (max_width[2] - len(table_row[2]) + 7))
            self.ptdebug(f"{table_row[3]}", end=" " * (max_width[3] - len(table_row[3]) + 2))
            self.ptdebug(" ")
        self.ptdebug("\n" + result["title"])
        self.ptdebug(f"\n{result['listed_info']}\n")



from enum import Enum
import dns.resolver
import dns.rdataclass
import dns.message
import dns.query
import dns.zone
import dns.dnssec
import whois 
import argparse
from typing import List, NamedTuple, Optional
from dataclasses import dataclass
from concurrent.futures import ThreadPoolExecutor, as_completed

from ptlibs.ptjsonlib import PtJsonLib

from ._base import BaseModule, BaseArgs, Out
from .utils.helpers import text_or_file

class VULNS(Enum):
    ZoneTransfer = "PTV-DNS-ZONETRANSFER"
    Subdomains = "PTV-DNS-SUBDOMAINS"
    DNSSEC = "PTV-DNS-DNSSEC"
    ZoneWalk = "PTV-DNS-ZONEWALK"


class info(NamedTuple):
    IP: str | None
    NSID: str | None
    bind_version: str | None   

@dataclass
class DNSResult:
    Info: Optional[List[info]] = None
    ReverseDomain: Optional[dict[str, list[str]]] = None
    ZoneTransfer: Optional[list[str]] = None
    Records: Optional[dict[str, dict[str, list[str]]]] = None
    Whois: Optional[dict[str, str]] = None
    Subdomains:Optional[list[str]] = None
    DNSSEC:Optional[list[str]]= None
    Zonewalk:Optional[list[str]] = None
    Zonewalk_com:Optional[list[str]] = None


class DNSArgs(BaseArgs):
    ip: str 
    port: int
    domain: str 
    subdomains: str 
    ip_file: str
    domain_file:str
    output:str
    command: str
    threads: int

    @staticmethod
    def get_help():
        return [
            {"description": ["DNS Testing Module"]},
            {"usage": ["ptsrvtester dns <command> <options>"]},
            {"usage_example": [
                "ptsrvtester dns whois -d example.com",
                "ptsrvtester dns lookup -d example.com",
                "ptsrvtester dns zone-transfer -d example.com"
            ]},
            {"options": [
                ["info", "<options>", "", "Retrieve DNS server information"],
                ["reverse-dns", "<options>", "", "Perform reverse DNS lookup"],
                ["zone-transfer", "<options>", "", "Attempt DNS zone transfer"],
                ["lookup", "<options>", "", "Query and display DNS records"],
                ["whois", "<options>", "", "Perform WHOIS lookup"],
                ["brute-subdomains", "<options>", "", "Brute-force subdomain enumeration"],
                ["dnssec", "<options>", "", "Check DNSSEC status"],
                ["zone-walk", "<options>", "", "DNS zone walking using NSEC/NSEC3"],
                ["zone-walk-complete", "<options>", "", "Complete zone walking using NSEC"],
                ["", "", "", ""],
                ["-h", "--help", "", "Show this help message and exit"],
            ]}
        ]

    def add_subparser(self, name: str, subparsers) -> None:
        """Adds a subparser of SNMP arguments"""

        examples = """example usage:
    ptsrvtester dns whois -d example.com 
    ptsrvtester dns lookup -d example.com
    """
        parser = subparsers.add_parser(
            name,
            epilog=examples,
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        if not isinstance(parser, argparse.ArgumentParser):
            raise TypeError

        dns_subparsers = parser.add_subparsers(dest="command", help="Select DNS command", required=True)

        # DNS info
        dns_info = dns_subparsers.add_parser("info", help="Retrieve DNS server information, including BIND version, NSID, and id.server.")
        dns_info.add_argument("-ip", "--ip", help="IP address of the target DNS server.")
        dns_info.add_argument("-p", "--port", type=int, default=53, help="Port of the DNS server (default: 53).")
        dns_info.add_argument("-ips", "--ip_file", help="File containing a list of target DNS server IP addresses.")

        # Reverse DNS Lookup
        reverse_dns_parser = dns_subparsers.add_parser("reverse-dns", help="Perform a reverse DNS lookup to resolve domain names from IP addresses.")
        reverse_dns_parser.add_argument("-ip", "--ip", help="IP address to perform reverse lookup.")
        reverse_dns_parser.add_argument("-ips", "--ip_file", help="File containing a list of target DNS server IP addresses.")
        
        # Zone Transfer
        zone_transfer_parser = dns_subparsers.add_parser("zone-transfer", help="Attempt a DNS zone transfer to enumerate DNS records from authoritative servers.")
        zone_transfer_parser.add_argument("-d", "--domain", required=True, help="Domain to attempt zone transfer.")
        zone_transfer_parser.add_argument("-o", "--output", help="File to save the output results.")
        
        # DNS Lookup 
        lookup_parser = dns_subparsers.add_parser("lookup", help="Query and display specified DNS records.")
        lookup_parser.add_argument("-d", "--domain", help="Domain name to analyze.")
        lookup_parser.add_argument("-rec", "--lookup-records", nargs='+', help="Specify DNS record types to query (default: A, AAAA, MX, TXT, CNAME, NS, SRV, PTR, SOA).")
        lookup_parser.add_argument("-dl", "--domain_file", help="File containing a list of domains to query.")
        
        # WHOIS Lookup
        whois_parser = dns_subparsers.add_parser("whois", help="Perform a WHOIS lookup to retrieve domain registration details.")
        whois_parser.add_argument("-d", "--domain", help="Domain name to analyze.")
        whois_parser.add_argument("-o", "--output", help="File to save the output results.")
        whois_parser.add_argument("-dl", "--domain_file", help="File containing a list of domains to analyze.")

        # Brute-force Subdomain Enumeration
        brute_subdomains_parser = dns_subparsers.add_parser("brute-subdomains", help="Conduct a brute-force attack to discover subdomains using a specified wordlist.")
        brute_subdomains_parser.add_argument("-d", "--domain", required=True, help="Target domain.")
        brute_subdomains_parser.add_argument("-sub", "--subdomains", required=True, help="Path to subdomains wordlist.")
        brute_subdomains_parser.add_argument("-o", "--output", help="File to save the output results.")
        brute_subdomains_parser.add_argument("-dl", "--domain_file", help="File containing a list of domains to analyze.")
        brute_subdomains_parser.add_argument("--threads", type=int, default=10, help="Number of threads to use for brute-force (default: 10)")
        # DNSSEC Check
        dnssec_parser = dns_subparsers.add_parser("dnssec", help="Check if the target domain has DNSSEC enabled and properly configured.")
        dnssec_parser.add_argument("-d", "--domain", help="Domain to check DNSSEC status.")
        dnssec_parser.add_argument("-dl", "--domain_file", help="File containing a list of domains to analyze.")

        # Zone Walking
        zone_walk_parser = dns_subparsers.add_parser("zone-walk", help="Attempt DNS zone walking using NSEC/NSEC3 records to enumerate available subdomains.")
        zone_walk_parser.add_argument("-d", "--domain", required=True, help="Domain to perform zone walking.")
        zone_walk_parser.add_argument("-o", "--output", help="File to save the output results.")
        
        # Complete Zone Walking
        zone_walk_complete_parser = dns_subparsers.add_parser("zone-walk-complete", help="Perform a full zone walking attempt to enumerate all subdomains using NSEC records.")
        zone_walk_complete_parser.add_argument("-d", "--domain", required=True, help="Domain to perform complete zone walking.")
        zone_walk_complete_parser.add_argument("-o", "--output", help="File to save the output results.")

class DNS(BaseModule):

    @staticmethod
    def module_args():
        return DNSArgs()

    def __init__(self, args: BaseArgs, ptjsonlib: PtJsonLib):
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.results: DNSResult | None = None

    def run(self) -> None:

        self.results = DNSResult()

        if self.args.command == "info":
            self.results.Info = self.print_dns_info()

        elif self.args.command == "reverse-dns":
            self.results.ReverseDomain = self.reverseDNS()

        elif self.args.command == "zone-transfer":
            self.results.ZoneTransfer = self.zone_transfer()

        elif self.args.command == "lookup":
            self.results.Records = self.lookup_dns_records(self.args.lookup_records)

        elif self.args.command == "whois":
            self.results.Whois = self.lookup_whois()

        elif self.args.command == "brute-subdomains":
            self.results.Subdomains = self.brute_force_subdomains()

        elif self.args.command == "dnssec":
            self.results.DNSSEC = self.check_dns()

        elif self.args.command == "zone-walk":
            self.results.Zonewalk = self.zone_walking()

        elif self.args.command == "zone-walk-complete":
            self.results.Zonewalk_com = self.zone_walking_complete()

        else:
            self.ptprint("Unknown command for DNS module.", out=Out.WARNING)


    def drawLine(self):
        self.ptprint('-' * 75)

    def drawDoubleLine(self):
        self.ptprint('=' * 75)

    def write_to_file(self, message_or_messages: str | list[str]):
        """
            File Output.
        """
        try:
            with open(self.args.output, "a") as f:
                if isinstance(message_or_messages, str):
                    f.write(message_or_messages + "\n")
                elif isinstance(message_or_messages, list):
                    for message in message_or_messages:
                        f.write(message + "\n")
        except FileNotFoundError:
            raise argparse.ArgumentError(None, f"File not found: '{self.args.output}'")
        except PermissionError:
            raise argparse.ArgumentError(
                None, f"Cannot write file (permission denied): '{self.args.output}'"
            )
        except OSError as e:
            raise argparse.ArgumentError(None, f"Cannot write file '{self.args.output}': {e}")

    def _text_or_file(self, text: str | None, file_path: str | None):
        """
            One domain/address or file.
        """
        values = text_or_file(text.strip() if text else None, file_path)
        return [v.strip() for v in values if v.strip()]


    def get_version_bind(self):
        """
            Retrieve the BIND version from the DNS server.
        """
        ip_addrs = self._text_or_file(self.args.ip, self.args.ip_file)
        results = {}
        
        for ip in ip_addrs:
            resolver = dns.resolver.Resolver()
            resolver.nameservers = [ip]
            
            try:
                answer = resolver.resolve('version.bind', 'TXT', rdclass=dns.rdataclass.CH)
                results[ip] = [record.to_text() for record in answer]
            except Exception as e:
                results[ip] =f"Error: {e}"
        
        return results
    
    def get_id_server_and_nsid(self):
        """ 
            Retrieve id.server TXT record and NSID from the DNS server. 
        """
        query = dns.message.make_query("id.server", dns.rdatatype.TXT, rdclass=dns.rdataclass.CH)
        query.use_edns(edns=0, ednsflags=0, options=[dns.edns.GenericOption(3, b"")])
        ip_addrs = self._text_or_file(self.args.ip, self.args.ip_file)
        results = {}
        
        for ip in ip_addrs:
            try:
                response = dns.query.udp(query, ip, port=self.args.port, timeout=5)
                id_server = next((item.to_text().strip('"') for answer in response.answer if answer.rdtype == dns.rdatatype.TXT
                                for item in answer.items), None)
                nsid = next((opt.nsid for opt in response.opt[0].options if opt), None) if response.opt else None
                results[ip] = (id_server, nsid)
            except Exception as e:
                results[ip] =f"Error: {e}"
        
        return results
    
    def print_dns_info(self) -> list[info]:
        """ 
            Print formatted DNS information. 
        """

        results: list[info] = []
        decoded_nsid = None
        
        self.drawDoubleLine()
        self.ptprint("Retrieving DNS banner information...", title=True)
        self.drawDoubleLine()
        self.ptprint("{:<20} {:<30}".format("Field", "Value"))
        self.drawLine()
        
        id_server_results = self.get_id_server_and_nsid()
        bind_version_results = self.get_version_bind()
        
        for ip, result in id_server_results.items():
            self.ptprint(f"Results for {ip}:")
            
            # Kontrola, zda je výsledek chybová zpráva
            if isinstance(result, str):
                self.ptprint(result, out=Out.ERROR)
            else:
                id_server, nsid = result
                decoded_nsid = nsid.decode() if nsid and isinstance(nsid, bytes) else None
                self.ptprint("{:<20} {:<30}".format("id.server", id_server if id_server else "Not available"))
                if nsid:
                    self.ptprint("{:<20} {:<30}".format("NSID (raw)", nsid.decode() if nsid and nsid.isascii() else nsid))
                    self.ptprint("{:<20} {:<30}".format("NSID (hex)", nsid.hex()))
            
            # Získání bind.version
            bind_version = bind_version_results.get(ip, ["Not available"])
            self.ptprint("{:<20} {:<30}".format("bind.version", ", ".join(bind_version) if isinstance(bind_version, list) else bind_version))
            self.drawLine()

            results.append(info(ip, decoded_nsid, bind_version))


    def reverseDNS(self) -> dict[str, list[str]]:
        """
            ipv4 addr -----> domain name
        """
        ip_addrs = self._text_or_file(self.args.ip, self.args.ip_file)
        results = {}

        for ip in ip_addrs:
            self.drawDoubleLine()
            self.ptprint(f"Performing reverse DNS lookup for {ip}...", title=True)
            self.drawDoubleLine()

            try:
                rev_name = dns.reversename.from_address(ip)  # Převod IP na reverzní formát
                response = dns.resolver.resolve(rev_name, "PTR")  # Dotaz na PTR záznam
                domains = [str(r) for r in response]  # Konverze výsledků na stringy
                answer = ', '.join(domains) if domains else "No PTR record"
                results[ip] = domains if domains else ["No PTR record"]
                self.ptprint("{:30} {:15}\n".format(ip, answer.rstrip('.')))
            except dns.resolver.NXDOMAIN:
                self.ptprint(f"Error: No PTR record found for {ip}", out=Out.ERROR)
                
            except dns.resolver.NoAnswer:
                self.ptprint(f"Error: No answer for PTR record query on {ip}", out=Out.ERROR)
            
            except dns.resolver.Timeout:
                self.ptprint(f"Error: Query timeout for {ip}", out=Out.WARNING)
           

        return results

    def resolveDNS(self, name):
        """
            domain name -----> ipv4 addr
        """
        resolver = dns.resolver.Resolver()
        try:
            results = resolver.resolve(name, "A")
            return results
        except dns.resolver.NoAnswer:
            self.ptprint(f"{name:<50} No A record found")
            return []
        

    def getNS (self, domain):
        """
            Name serververs for the domain.
        """
        mapping = {}
        try:
            name_servers = dns.resolver.resolve(domain, 'NS')
            for name_server in name_servers:
                A_records = self.resolveDNS(str(name_server))
                if A_records:
                    for item in A_records:
                        answer = ','.join([str(item)])
                    mapping[str(name_server)] = answer
                    self.ptprint("{:<50} {:<15}".format(str(name_server).rstrip('.'), answer))
                else:
                    mapping[str(name_server)] = "No A record"
                    self.ptprint("{:<50} {:<15}".format(str(name_server).rstrip('.'), "No A record"))
            return mapping
        
        except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer, dns.resolver.Timeout) as e:
            self.ptprint(f"Error: Unable to retrieve authoritative NS - {e}", out=Out.ERROR)
            return None
    
    

    def zoneXFR(self, server):
        """
            Zone transfer from the name server
        """
        found_domains = []

        try:
            zone = dns.zone.from_xfr(dns.query.xfr(str(server).rstrip('.'), self.args.domain))
        except Exception as e:
            self.ptprint(f"Error: {e.__class__} {e}", out=Out.ERROR)
        else:
            for host in zone:
                if str(host) != '@':
                    A_records = self.resolveDNS(str(host) + "." + self.args.domain)
                    if A_records: 
                        for item in A_records:
                            answer = ','.join([str(item)]) 
                        domain_entry = f"{host}.{self.args.domain};{answer}"
                        found_domains.append(domain_entry)
                        fulldomain = f"{str(host)}.{self.args.domain}"
                        self.ptprint(f"{fulldomain:<50} {answer}")
                    else:
                        found_domains.append(f"{host}.{self.args.domain}")
                else:
                    found_domains.append(f"{host}.{self.args.domain}")
            return found_domains


    def zone_transfer(self) -> list[str]:
        """
            Zone transfer formated output

            returns list of nameserver which allows zone transfer
        """
        result = []
        any_domains_found = False

        self.drawDoubleLine()
        self.ptprint(f"Retrieving name servers for {self.args.domain}...", title=True)
        self.drawDoubleLine()
        name_servers = self.getNS(self.args.domain)
        self.drawLine()
        if name_servers == None:
            self.ptprint(f"Error: Unable to start zone transfer for {self.args.domain}", out=Out.ERROR)
        else:
            for server in name_servers:
                self.ptprint("\n")
                self.ptprint(f"Attempting zone transfer from {server} ({name_servers[server]})...", out=Out.TITLE)
                self.drawLine()
                found_domains = self.zoneXFR(name_servers[server])
                if self.args.output:
                    self.write_to_file(found_domains)
                if found_domains:
                     result.append(server) 
                  
        return result



    def lookup_dns_records(self, record_types=None, banner = True) ->  dict[str, dict[str, list[str]]]:
        """
             Retrieve DNS records from the domain.
        """
        results = {}
        if record_types is None:
            # Default list
            record_types = ["A", "AAAA", "MX", "TXT", "CNAME", "NS", "SRV", "PTR", "SOA"]
        
        domains = self._text_or_file(self.args.domain, self.args.domain_file)
        for domain in domains:

            if banner:
                self.drawDoubleLine()
                self.ptprint(f"Querying DNS records for {domain}...", title=True)
                self.drawDoubleLine()
            self.ptprint("{:<15} {:<30}".format("Record type", "Value"))
            self.drawLine()

            for record_type in record_types:
                try:
                    answers = dns.resolver.resolve(domain, record_type)
                    records = [str(r) for r in answers]

                    if domain not in results:
                        results[domain] = {}
                    results[domain][record_type] = records

                    self.ptprint("{:<15} {:<30}".format(record_type, ", ".join(records)))
                except dns.resolver.NXDOMAIN:
                    self.ptprint(f"{record_type:<15} No records found (NXDOMAIN)")
                except dns.resolver.NoAnswer:
                    self.ptprint(f"{record_type:<15} No answer")
                except dns.resolver.Timeout:
                    self.ptprint(f"{record_type:<15} Query timeout")
                except dns.resolver.NoNameservers:
                    self.ptprint(f"{record_type:<15} No available name servers")
                except Exception as e:
                    self.ptprint(f"{record_type:<15} Error - {e}")

        return results
    
    def lookup_whois(self) -> dict[str, str]:
        """
            Perform a WHOIS lookup for a given domain.
        """
        results = {}
        domains = self._text_or_file(self.args.domain, self.args.domain_file)
        for domain in domains:
            self.drawDoubleLine()
            self.ptprint(f"Querying WHOIS information for {domain}...", title=True)
            self.drawDoubleLine()

            try:
                info = whois.whois(domain)  # all whois records
                whois_text = info.text if hasattr(info, "text") else str(info) 
                self.ptprint(whois_text)

                results[domain] = whois_text
                
                if self.args.output:
                    self.write_to_file(whois_text)

            except Exception as e:
                self.ptprint(f"Error: {e}", out=Out.WARNING)
                
        return results

    def brute_force_subdomains(self) -> list[str]:
        """
            Perform a brute-force attack to find subdomains using a wordlist.
        """
        domains = self._text_or_file(self.args.domain, self.args.domain_file)
        for domain in domains:
            self.drawDoubleLine()
            self.ptprint(f"Starting brute-force attack on subdomains for {domain}...", title=True)
            self.drawDoubleLine()

            found_subdomains = []
            output = []
            record_types = ["A", "AAAA", "MX", "TXT", "CNAME", "NS", "SRV", "PTR", "SOA"]

            try:
                subdomains = [s.strip() for s in text_or_file(None, self.args.subdomains) if s.strip()]

                def resolve_subdomain(sub: str):
                    subdomain = f"{sub}.{domain}"
                    subdomain_records = {}
                    for record_type in record_types:
                        try:
                            answers = dns.resolver.resolve(subdomain, record_type)
                            subdomain_records[record_type] = [str(r) for r in answers]
                        except Exception:
                            continue
                    if subdomain_records:
                        return (subdomain, subdomain_records)
                    return None

                with ThreadPoolExecutor(max_workers=self.args.threads) as executor:
                    futures = [executor.submit(resolve_subdomain, sub) for sub in subdomains]
                    for future in as_completed(futures):
                        result = future.result()
                        if result:
                            subdomain, records = result
                            found_subdomains.append((subdomain, records))
                            output.append(subdomain)
                            self.ptprint(f"Found: {subdomain}", out=Out.OK)
                            for rtype, values in records.items():
                                self.ptprint(f"   └── {rtype}: {', '.join(values)}")

                self.ptprint("\n")
                self.ptprint(f"Brute-force attack completed. Found {len(found_subdomains)} subdomains.", out=Out.INFO)
                if self.args.output:
                    self.write_to_file(output)

                results = [sub[0] for sub in found_subdomains]
                return results

            except argparse.ArgumentError:
                raise
        
    def get_nsip(self, domain):
        """
            Get authoritatve nameserver for the domain and its ip address
        """
        try:
            # Get authoritative nameservers for the domain
            ns_response = dns.resolver.resolve(domain, dns.rdatatype.NS)
            if not ns_response.rrset or len(ns_response.rrset) == 0:
                self.ptprint(f"Error: No NS records found for domain {domain}", out=Out.ERROR)
                return None
            
            ns_name = str(ns_response.rrset[0])

            # Use resolveDNS() to get the NS IP address
            ns_ip_results = self.resolveDNS(ns_name)
            if not ns_ip_results:
                self.ptprint(f"Error: No IP found for NS {ns_name}", out=Out.WARNING)
                return None

            else:
                ns_ip = str(ns_ip_results[0])
                return ns_ip

        except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer, dns.resolver.Timeout) as e:
            self.ptprint(f"Error: Unable to retrieve authoritative NS - {e}", out=Out.ERROR)
            return None

    
    def check_dns(self) -> list[str] | None:
        """
            Check if the domain has DNSSEC.
        """
        insecure_domains = []
        domains = self._text_or_file(self.args.domain, self.args.domain_file)
        for domain in domains:
            self.drawDoubleLine()
            self.ptprint(f"Checking DNSSEC on {domain}...", title=True)
            self.drawDoubleLine()

            domain = domain.rstrip(".") + "."
            ns_ip = self.get_nsip(domain)

            try:
                # Query for DNSKEY records with DNSSEC support
                request = dns.message.make_query(domain, dns.rdatatype.DNSKEY, want_dnssec=True)
                response = dns.query.udp(request, ns_ip, timeout=10)

                if response.rcode() != 0:
                    self.ptprint("Error: No DNSKEY record found", out=Out.VULN)
                    insecure_domains.append(domain)
                    

                # Check if the response contains DNSKEY and RRSIG(DNSKEY)
                if len(response.answer) < 2:
                    self.ptprint("DNSSEC is NOT properly set! (Missing DNSSEC signatures RRSIG/DNSKEY)", out=Out.VULN)
                    self.ptprint("Status: Insecure", out=Out.VULN)
                    

                # Extract DNSKEY and RRSIG records
                dnskey_rrset = response.answer[0]
                rrsig_rrset = response.answer[1]

                # Validate the self-signed DNSKEY
                name = dns.name.from_text(domain)
                dns.dnssec.validate(dnskey_rrset, rrsig_rrset, {name: dnskey_rrset})

                self.ptprint("DNSSEC is ENABLED and properly signed!", out=Out.NOTVULN)
                self.ptprint("Status: Secure", out=Out.NOTVULN)
                self.ptprint("\n")
                self.ptprint("DNSKEY Details:", out=Out.INFO)
                self.drawLine()

                for key in dnskey_rrset:
                    key_id = dns.dnssec.key_id(key)  # Extract Key ID (Key Tag)
                    algorithm = dns.dnssec.algorithm_to_text(key.algorithm)  # Get signing algorithm
                    key_type = "KSK" if key.flags == 257 else "ZSK"  # 257 = KSK, 256 = ZSK

                    self.ptprint(f"{'Key ID:':<20} {key_id:<30}")
                    self.ptprint(f"{'Algorithm:':<20} {algorithm:<30}")
                    self.ptprint(f"{'Key Type:':<20} {key_type:<30}\n")

            except dns.dnssec.ValidationFailure:
                self.ptprint("DNSSEC validation failed! Domain is NOT properly signed.", out=Out.VULN)
                self.ptprint("Status: Insecure", out=Out.VULN)
                insecure_domains.append(domain)
                
            except dns.exception.Timeout:
                self.ptprint("Error: DNSSEC validation timeout.", out=Out.WARNING)
                insecure_domains.append(domain)
                
            except Exception as e:
                self.ptprint(f"Error: Unexpected error - {e}", out=Out.WARNING)

        return insecure_domains
                

    def zone_walking(self) -> list[str] | None:
        """
            Attempts to enumerate subdomains using NSEC/NSEC3 DNSSEC records.
        """

        self.drawDoubleLine()
        self.ptprint(f"Starting zone walking for {self.args.domain} using NSEC/NSEC3", title=True)
        self.drawDoubleLine()

        domain = self.args.domain.rstrip(".") + "."

        ns_ip = self.get_nsip(domain)


        # Means that the given domain doesnt exist
        if ns_ip == None:
            self.drawLine()
            split_domain = domain.split(".")

            for i in range(1, len(split_domain) - 1):  # Start from parent domains
                parent_domain = ".".join(split_domain[i:])
                ns_ip = self.get_nsip(parent_domain)

                if ns_ip != None:
                    break

        try:
            # Query for NSEC/NSEC3 records
            request = dns.message.make_query(domain, dns.rdatatype.NSEC, want_dnssec=True)
            response = dns.query.udp(request, ns_ip, timeout=10)
 
            found_subdomains = []    #NSEC
            hashed_subdomains = []   #NSEC3

            for rrset in response.authority:
                #NSEC
                if rrset.rdtype == dns.rdatatype.NSEC:
                    owner = rrset.name.to_text().rstrip(".")  # Current domain in the NSEC chain
                    found_subdomains.append(owner)  # Store the owner name (existing domain)
                    next_domain = rrset[0].next.to_text().rstrip(".")
                    found_subdomains.append(next_domain)  # Store next domain

                #NSEC3
                if rrset.rdtype == dns.rdatatype.NSEC3:
                    for rrset in response.authority:
                        domain_hash = rrset.name.to_text().rstrip(".")  # Extract the NSEC3 hashed domain
                        hashed_subdomains.append(domain_hash)  # Add to set to avoid duplicates

            self.ptprint("\n")
            self.ptprint(f"Zone walking results using NSEC/NSEC3 records:", out=Out.INFO)
            self.drawLine()

            if found_subdomains:
                if self.args.output: 
                    self.write_to_file(list(dict.fromkeys(found_subdomains)))
                for sub in dict.fromkeys(found_subdomains):
                    self.ptprint(f"{sub}")
                found_subdomains = list(dict.fromkeys(found_subdomains))
                return found_subdomains

            elif hashed_subdomains:
                if self.args.output: 
                    self.write_to_file(list(dict.fromkeys(hashed_subdomains)))
                for hashed in dict.fromkeys(hashed_subdomains):
                    self.ptprint(f"{hashed}")
                hashed_subdomains=list(dict.fromkeys(hashed_subdomains))
                return hashed_subdomains

            else: 
                self.ptprint("No subdomains discovered.", out=Out.ERROR)
                return None 
        
        except dns.exception.Timeout:
            self.ptprint("Error: Query timeout for NSEC enumeration.", out=Out.WARNING)
            return None
        except Exception as e:
            self.ptprint(f"Error: Unexpected error - {e}", out=Out.WARNING)
            return None
        

    def zone_walking_complete(self) -> list[str] | None:
        """
            Attempts to enumerate subdomains using NSEC.
        """
        
        self.drawDoubleLine()
        self.ptprint(f"Starting complete zone walking using NSEC from {self.args.domain}", title=True)
        self.drawDoubleLine()

        domain = self.args.domain.rstrip(".") + "."
        ns_ip = self.get_nsip(domain)


        # Means that the given domain doesnt exist
        if ns_ip == None:
            self.drawLine()
            split_domain = domain.split(".")

            for i in range(1, len(split_domain) - 1):  # Start from parent domains
                parent_domain = ".".join(split_domain[i:])
                ns_ip = self.get_nsip(parent_domain)

                if ns_ip != None:
                    break

        queue = [domain]  # Start with the given domain
        found_subdomains = []
        count = 0
        

        try:
            while queue and count != 3:
                last_len = len(found_subdomains)
                current_domain = queue.pop(0)

                # Query for NSEC records
                request = dns.message.make_query(current_domain, dns.rdatatype.NSEC, want_dnssec=True)
                response = dns.query.udp(request, ns_ip, timeout=10)

                for rrset in response.authority:
                    #NSEC
                    if rrset.rdtype == dns.rdatatype.NSEC:
                        owner = rrset.name.to_text().rstrip(".")  # Current domain in the NSEC chain
                        next_domain = rrset[0].next.to_text().rstrip(".")

                        if owner not in found_subdomains:
                            found_subdomains.append(owner)

                        if next_domain and next_domain not in found_subdomains:
                            found_subdomains.append(next_domain)

                        next_domain = "0." + next_domain
                        queue.append(next_domain)  # Queue next domain for walking
                        current_len = len(found_subdomains)

                        if current_len == last_len:
                            count += 1
                        else:
                            count = 0

            self.ptprint("\n")
            self.ptprint(f"Complete zone walking results using NSEC records:", out=Out.INFO)
            self.drawLine()

            unique_subs = list(dict.fromkeys(found_subdomains)) #to awoid duplicities

            if found_subdomains:
                for sub in sorted(unique_subs):
                    self.ptprint(sub)
                if self.args.output: 
                    self.write_to_file(unique_subs)
                return unique_subs

            else: 
                self.ptprint("No subdomains discovered.", out=Out.ERROR)
                return None 
        
        except dns.exception.Timeout:
            self.ptprint("Error: Query timeout for NSEC enumeration.", out=Out.WARNING)
            return None
        except Exception as e:
            self.ptprint(f"Error: Unexpected error - {e}", out=Out.WARNING)
            return None
        
    def output(self) -> None:
        """
        Info: Optional[List[info]] = None
        ReverseDomain: Optional[dict[str, list[str]]] = None
        ZoneTransfer: Optional[list[str]] = None
        Records: Optional[dict[str, list[str]]] = None
        Whois: Optional[dict[str, str]] = None
        Subdomains:Optional[list[str]] = None
        DNSSEC:Optional[bool]= None
        Zonewalk:Optional[list[str]] = None
        Zonewalk_com:Optional[list[str]] = None

        ZoneTransfer = "PTV-SNMPv2-ZONETRANFER"
        Subdomains = "PTV-SNMPv3-SUBDOMAINS"
        DNSSEC = "PTV-SNMPv2-DNSSEC"
        ZoneWalk = "PTV-SNMPv3-ZONEWALK"
        """
        properties = {
            "software_type": None,
            "name": "dns",
            "version": None,
            "vendor": None,
            "description": None,
            "zoneTransfer": self.results.ZoneTransfer if self.results.ZoneTransfer is not None else None,
            "subdomains": self.results.Subdomains if self.results.Subdomains is not None else None,
            "dnssec": self.results.DNSSEC if self.results.DNSSEC is not None else None,
            "zonewalk": self.results.Zonewalk if self.results.Zonewalk is not None else None,
        }
        deferred_vulns = []

        if self.results.ZoneTransfer is not None and len(self.results.ZoneTransfer) != 0:
            deferred_vulns.append({
                "vuln_code": VULNS.ZoneTransfer.value,
                "vuln_request": "Testing for Zone transfer",
                "vuln_response": ",".join(self.results.ZoneTransfer),
            })
        if self.results.Subdomains is not None and len(self.results.Subdomains) != 0:
            deferred_vulns.append({
                "vuln_code": VULNS.Subdomains.value,
                "vuln_request": "Subdomains discovered via brute-force enumeration",
                "vuln_response": ",".join(self.results.Subdomains),
            })
        if self.results.DNSSEC is not None and len(self.results.DNSSEC) != 0:
            deferred_vulns.append({
                "vuln_code": VULNS.DNSSEC.value,
                "vuln_request": "Domains without DNSSEC or with invalid DNSSEC configuration",
                "vuln_response": ",".join(self.results.DNSSEC),
            })
        if self.results.Zonewalk is not None and len(self.results.Zonewalk) != 0:
            deferred_vulns.append({
                "vuln_code": VULNS.ZoneWalk.value,
                "vuln_request": "Subdomains discovered via DNS zone walking (NSEC/NSEC3)",
                "vuln_response": ",".join(self.results.Zonewalk),
            })

        dns_node = self.ptjsonlib.create_node_object("software", None, None, properties)
        self.ptjsonlib.add_node(dns_node)
        node_key = dns_node["key"]
        for v in deferred_vulns:
            self.ptjsonlib.add_vulnerability(node_key=node_key, **v)

        self.ptprint(self.ptjsonlib.get_result_json(), json=True)


"""
Service identification from server banner.
Maps banner content to product name, version (if numeric), OS/distribution, and CPE 2.3 (NVD-style).
Only numeric app versions (e.g. 7.1.1, 4.96) are used as version and in CPE; distribution
names (e.g. Ubuntu, Debian/GNU Linux) from banner go to os, not version.

CPE 2.3 (NIST/NVD) has exactly 11 components after the "cpe:2.3:" prefix:
  part, vendor, product, version, update, edition, language, sw_edition, target_sw, target_hw, other.
Unknown/NA components are represented as *.
"""
import re
from dataclasses import dataclass
from typing import Final

# Version string acceptable for CPE (NVD uses numeric versions). Reject distro names.
_CPE_VERSION_RE = re.compile(r"^\d+[.\d]*(?:p\d+)?[\w.-]*$", re.I)

# CPE 2.3: exactly 11 components after "cpe:2.3:" (version, update, edition, language, sw_edition, target_sw, target_hw, other)
_CPE_23_NUM_COMPONENTS = 11
# Wildcard tail when version is unknown (8 fields after vendor:product)
_CPE_23_ANY_TAIL = ":*:*:*:*:*:*:*:*"


def _cpe_23_normalize(cpe: str) -> str:
    """Return CPE 2.3 string with all 11 components; pad with * if shorter."""
    if not cpe or not cpe.startswith("cpe:2.3:"):
        return cpe
    parts = cpe.split(":")
    if len(parts) < 3:
        return cpe
    # parts[0]=cpe, parts[1]=2.3, parts[2:]=component list
    components = parts[2:]
    while len(components) < _CPE_23_NUM_COMPONENTS:
        components.append("*")
    return ":".join([parts[0], parts[1]] + components[: _CPE_23_NUM_COMPONENTS])

# (pattern re, product display name, CPE base with wildcard version … _CPE_23_ANY_TAIL)
# Order: more specific first. Version is first capture group if present.
# CPE vendor/product from NVD where available.
_BANNER_PATTERNS: Final[list[tuple[re.Pattern[str], str, str]]] = [
    # --- SMTP / mail (MTAs, gateways) ---
    (re.compile(r"Kerio\s+Connect\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Kerio Connect", "cpe:2.3:a:kerio:connect:*:*:*:*:*:*:*:*"),
    (re.compile(r"Microsoft\s+ESMTP\s+MAIL\s+Service", re.I), "Microsoft Exchange Server", "cpe:2.3:a:microsoft:exchange_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"Microsoft\s+SMTP\s+Server", re.I), "Microsoft SMTP Server", "cpe:2.3:a:microsoft:exchange_server:*:*:*:*:*:*:*:*"),
    # Cloud providers first – banner is authoritative (similar EHLO across providers)
    (re.compile(r"mx\.google\.com|smtp\.google\.com|\bgsmtp\b", re.I), "Google Workspace", "cpe:2.3:a:google:gmail:*:*:*:*:*:*:*:*"),
    (re.compile(r"outlook\.com|smtp\.office365\.com|protection\.outlook\.com", re.I), "Microsoft 365", "cpe:2.3:a:microsoft:exchange_online:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bzoho\b", re.I), "Zoho Mail", "cpe:2.3:a:zoho:mail:*:*:*:*:*:*:*:*"),
    (re.compile(r"mail\.protonmail\.ch|protonmail\.(ch|me|com)", re.I), "Proton Mail", "cpe:2.3:a:protonmail:protonmail:*:*:*:*:*:*:*:*"),
    (re.compile(r"mx\.messagingengine\.com", re.I), "Fastmail", "cpe:2.3:a:fastmail:fastmail:*:*:*:*:*:*:*:*"),
    (re.compile(r"mx\.yandex\.(ru|net|com)", re.I), "Yandex Mail", "cpe:2.3:a:yandex:yandex_mail:*:*:*:*:*:*:*:*"),
    (re.compile(r"yahoo\.com\s+ESMTP\s+ready", re.I), "Yahoo Mail Proxy", "cpe:2.3:a:yahoo:mail_proxy:*:*:*:*:*:*:*:*"),
    (re.compile(r"aol\.mail.*ESMTP\s+ready", re.I), "Yahoo Mail Proxy", "cpe:2.3:a:yahoo:mail_proxy:*:*:*:*:*:*:*:*"),
    (re.compile(r"Mailgun\s+Delivery\s+Service", re.I), "Mailgun", "cpe:2.3:a:mailgun:mailgun:*:*:*:*:*:*:*:*"),
    (re.compile(r"SendGrid\s+ESMTP", re.I), "SendGrid", "cpe:2.3:a:sendgrid:sendgrid:*:*:*:*:*:*:*:*"),
    (re.compile(r"email-smtp\.[a-z0-9-]*\.?amazonaws\.com\s+ESMTP", re.I), "Amazon SES", "cpe:2.3:a:amazon:ses:*:*:*:*:*:*:*:*"),
    (re.compile(r"mandrillapp\.com\s+ESMTP", re.I), "Mandrill (Mailchimp)", "cpe:2.3:a:mailchimp:mandrill:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+Sendinblue\s+SMTP\s+(\d+\.\d+)", re.I), "Sendinblue SMTP (Brevo)", "cpe:2.3:a:brevo:smtp:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bSendinblue\b", re.I), "Sendinblue SMTP (Brevo)", "cpe:2.3:a:brevo:smtp:*:*:*:*:*:*:*:*"),
    (re.compile(r"Cisco\s+Secure\s+Email", re.I), "Cisco Secure Email (IronPort)", "cpe:2.3:h:cisco:secure_email_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"(?:AsyncOS|IronPort)\s+(\d+\.\d+)", re.I), "Cisco Secure Email (IronPort)", "cpe:2.3:h:cisco:secure_email_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"\b(?:AsyncOS|IronPort)\b", re.I), "Cisco Secure Email (IronPort)", "cpe:2.3:h:cisco:secure_email_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+IceWarp\s+(\d+(?:\.\d+)+)", re.I), "IceWarp", "cpe:2.3:a:icewarp:icewarp_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bIceWarp\s+(\d+(?:\.\d+)+)", re.I), "IceWarp", "cpe:2.3:a:icewarp:icewarp_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bIceWarp\b", re.I), "IceWarp", "cpe:2.3:a:icewarp:icewarp_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+FortiMail\s+(\d+\.\d+(?:\.\d+)?)", re.I), "FortiMail", "cpe:2.3:h:fortinet:fortimail:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bFortiMail\b", re.I), "FortiMail", "cpe:2.3:h:fortinet:fortimail:*:*:*:*:*:*:*:*"),
    (re.compile(r"Oracle\s+Communications\s+Messaging\s+Server\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Oracle Communications Messaging Server", "cpe:2.3:a:oracle:communications_messaging_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"Oracle\s+Communications\s+Messaging\s+Server", re.I), "Oracle Communications Messaging Server", "cpe:2.3:a:oracle:communications_messaging_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"Zimbra\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Zimbra Collaboration", "cpe:2.3:a:zimbra:collaboration:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bZCS\b", re.I), "Zimbra Collaboration", "cpe:2.3:a:zimbra:collaboration:*:*:*:*:*:*:*:*"),
    (re.compile(r"hMailServer\s+(\d+\.\d+(?:\.\d+)?)", re.I), "hMailServer", "cpe:2.3:a:hmailserver:hmailserver:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bhMailServer\b", re.I), "hMailServer", "cpe:2.3:a:hmailserver:hmailserver:*:*:*:*:*:*:*:*"),
    (re.compile(r"Apache\s+James\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Apache James", "cpe:2.3:a:apache:james:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bJames\s+SMTP\b", re.I), "Apache James", "cpe:2.3:a:apache:james:*:*:*:*:*:*:*:*"),
    (re.compile(r"Haraka\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Haraka", "cpe:2.3:a:haraka:haraka:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bHaraka\b", re.I), "Haraka", "cpe:2.3:a:haraka:haraka:*:*:*:*:*:*:*:*"),
    (re.compile(r"Sendmail\s+(\d+\.\d+(?:\.\d+)?)(?:\/[^\s]*)?", re.I), "Sendmail", "cpe:2.3:a:sendmail:sendmail:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bSendmail\b", re.I), "Sendmail", "cpe:2.3:a:sendmail:sendmail:*:*:*:*:*:*:*:*"),
    (re.compile(r"Exim\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Exim", "cpe:2.3:a:exim:exim:*:*:*:*:*:*:*:*"),
    # Bare product name before generic "ESMTP service ready" (e.g. Hetzner: "host Exim ESMTP Service ready")
    (re.compile(r"\bExim\b", re.I), "Exim", "cpe:2.3:a:exim:exim:*:*:*:*:*:*:*:*"),
    (re.compile(r"Postfix\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Postfix", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    (re.compile(r"Postfix\s+\(([^)]+)\)", re.I), "Postfix", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+Postfix", re.I), "Postfix", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bPostfix\b", re.I), "Postfix", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    (re.compile(r"Lotus\s+Domino\s+(\d+\.\d+)", re.I), "IBM Lotus Domino", "cpe:2.3:a:ibm:domino:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bDomino\b", re.I), "IBM Lotus Domino", "cpe:2.3:a:ibm:domino:*:*:*:*:*:*:*:*"),
    (re.compile(r"MailEnable\s+Service,\s+Version:\s*(\d+\.\d+(?:\.\d+)?)", re.I), "MailEnable", "cpe:2.3:a:mailenable:mailenable:*:*:*:*:*:*:*:*"),
    (re.compile(r"MailEnable\s+(\d+\.\d+)", re.I), "MailEnable", "cpe:2.3:a:mailenable:mailenable:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMailEnable\b", re.I), "MailEnable", "cpe:2.3:a:mailenable:mailenable:*:*:*:*:*:*:*:*"),
    (re.compile(r"Axigen\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Axigen", "cpe:2.3:a:axigen:axigen_mail_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bAxigen\b", re.I), "Axigen", "cpe:2.3:a:axigen:axigen_mail_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"MDaemon\s+(\d+\.\d+)", re.I), "MDaemon", "cpe:2.3:a:alt-n:mdaemon:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMDaemon\b", re.I), "MDaemon", "cpe:2.3:a:alt-n:mdaemon:*:*:*:*:*:*:*:*"),
    (re.compile(r"SmarterMail\s+(\d+\.\d+(?:\.\d+)?)", re.I), "SmarterMail", "cpe:2.3:a:smartertools:smartermail:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bSmarterMail\b", re.I), "SmarterMail", "cpe:2.3:a:smartertools:smartermail:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+Postcow", re.I), "Mailcow: dockerized", "cpe:2.3:a:mailcow:mailcow:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bPostcow\b", re.I), "Mailcow: dockerized", "cpe:2.3:a:mailcow:mailcow:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMailcow\b", re.I), "Mailcow: dockerized", "cpe:2.3:a:mailcow:mailcow:*:*:*:*:*:*:*:*"),
    (re.compile(r"MailStore\s+Gateway", re.I), "MailStore Gateway", "cpe:2.3:a:mailstore:mailstore_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMailStore\b", re.I), "MailStore Gateway", "cpe:2.3:a:mailstore:mailstore_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"Barracuda\s+(\d+\.\d+)", re.I), "Barracuda Email Security", "cpe:2.3:h:barracuda:email_security_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bBarracuda\b", re.I), "Barracuda Email Security", "cpe:2.3:h:barracuda:email_security_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"Welcome\s+to\s+coremail", re.I), "Coremail Mail Server", "cpe:2.3:a:coremail:coremail:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bcoremail\b", re.I), "Coremail Mail Server", "cpe:2.3:a:coremail:coremail:*:*:*:*:*:*:*:*"),
    (re.compile(r"pphosted\.com", re.I), "Proofpoint Email Protection", "cpe:2.3:a:proofpoint:proofpoint_email_protection:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+mfa-", re.I), "Proofpoint Email Protection", "cpe:2.3:a:proofpoint:proofpoint_email_protection:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bmfa-\w+", re.I), "Proofpoint Email Protection", "cpe:2.3:a:proofpoint:proofpoint_email_protection:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bProofpoint\b", re.I), "Proofpoint Email Protection", "cpe:2.3:a:proofpoint:proofpoint_email_protection:*:*:*:*:*:*:*:*"),
    # Sophos: typical short banner before bracketed appliance form
    (re.compile(r"\bSophos\s+ESMTP\s+ready\b", re.I), "Sophos Email Appliance", "cpe:2.3:a:sophos:email_appliance:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+\[Sophos\s+Email\s+Appliance\]", re.I), "Sophos Email Appliance", "cpe:2.3:a:sophos:email_appliance:*:*:*:*:*:*:*:*"),
    (re.compile(r"Sophos\s+Email\s+Appliance", re.I), "Sophos Email Appliance", "cpe:2.3:a:sophos:email_appliance:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bSophos\b", re.I), "Sophos Email Appliance", "cpe:2.3:a:sophos:email_appliance:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+\(Mimecast\)", re.I), "Mimecast", "cpe:2.3:a:mimecast:email_security_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMimecast\b", re.I), "Mimecast", "cpe:2.3:a:mimecast:email_security_gateway:*:*:*:*:*:*:*:*"),
    (re.compile(r"ESMTP\s+Trend\s+Micro\s+Email\s+Security\s+Service\s+ready", re.I), "Trend Micro Email Security", "cpe:2.3:a:trendmicro:email_security:*:*:*:*:*:*:*:*"),
    (re.compile(r"Trend\s+Micro\s+Email\s+Security", re.I), "Trend Micro Email Security", "cpe:2.3:a:trendmicro:email_security:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bDovecot\b", re.I), "Dovecot", "cpe:2.3:a:dovecot:dovecot:*:*:*:*:*:*:*:*"),
    (re.compile(r"Courier\s+(?:mail\s+)?(?:server|Mail)", re.I), "Courier", "cpe:2.3:a:courier-mta:courier:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bQmail\b", re.I), "Qmail", "cpe:2.3:a:qmail:qmail:*:*:*:*:*:*:*:*"),
    # DreamHost VPS: dreamhostps.com hostname – typically runs Postfix
    (re.compile(r"dreamhostps\.com", re.I), "Postfix (DreamHost VPS)", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    # Etius.jp (WebArena): Japanese hosting provider – minimal banner, Postfix typical
    (re.compile(r"etius\.jp", re.I), "Postfix (Etius.jp / WebArena)", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    # LWS Hosting (lwspanel.com): French hosting – uses Postfix with "LWS mailer" banner
    (re.compile(r"LWS\s+mailer", re.I), "Postfix (LWS Hosting)", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    (re.compile(r"lwspanel\.com", re.I), "Postfix (LWS Hosting)", "cpe:2.3:a:postfix:postfix:*:*:*:*:*:*:*:*"),
    # PowerMTA (Port25): pmtaserver.com hostname – often masquerades as Exchange (custom ESMTP banner)
    (re.compile(r"pmtaserver\.com", re.I), "PowerMTA (Port25)", "cpe:2.3:a:port25:powermta:*:*:*:*:*:*:*:*"),
    # Generic "ESMTP service ready" only when no explicit MTA name matched above (Exim/Postfix/Sendmail/…)
    (re.compile(r"\bESMTP\s+service\s+ready\b", re.I), "Microsoft Exchange Server (custom banner)", "cpe:2.3:a:microsoft:exchange_server:*:*:*:*:*:*:*:*"),
    # --- IMAP / POP3 (Cyrus, etc.) ---
    (re.compile(r"Cyrus\s+POP3\s+v?(\d+\.\d+(?:\.\d+)?)", re.I), "Cyrus POP3", "cpe:2.3:a:cyrus:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"Cyrus\s+POP3", re.I), "Cyrus POP3", "cpe:2.3:a:cyrus:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"Cyrus\s+IMAP4?\s+v?(\d+\.\d+(?:\.\d+)?)", re.I), "Cyrus IMAP", "cpe:2.3:a:cyrus:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"Cyrus\s+IMAP", re.I), "Cyrus IMAP", "cpe:2.3:a:cyrus:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"University\s+of\s+Washington", re.I), "UW IMAP", "cpe:2.3:a:university_of_washington:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bUW\s+IMAP\b", re.I), "UW IMAP", "cpe:2.3:a:university_of_washington:imap:*:*:*:*:*:*:*:*"),
    (re.compile(r"CommuniGate\s+Pro\s+(\d+\.\d+)", re.I), "CommuniGate Pro", "cpe:2.3:a:stalker:communigate_pro:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bCommuniGate\b", re.I), "CommuniGate Pro", "cpe:2.3:a:stalker:communigate_pro:*:*:*:*:*:*:*:*"),
    # --- SSH ---
    (re.compile(r"SSH-2\.0-OpenSSH_([\d\.]+(?:p\d+)?)", re.I), "OpenSSH", "cpe:2.3:a:openbsd:openssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"OpenSSH_([\d\.]+(?:p\d+)?)", re.I), "OpenSSH", "cpe:2.3:a:openbsd:openssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-libssh_([\d\.]+)", re.I), "libssh", "cpe:2.3:a:libssh:libssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"libssh[_ ]([\d\.]+)", re.I), "libssh", "cpe:2.3:a:libssh:libssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-dropbear_(\d{4}\.\d+)", re.I), "Dropbear SSH", "cpe:2.3:a:dropbear_ssh_project:dropbear_ssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"dropbear[_ ](\d{4}\.\d+)", re.I), "Dropbear SSH", "cpe:2.3:a:dropbear_ssh_project:dropbear_ssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-dropbear", re.I), "Dropbear SSH", "cpe:2.3:a:dropbear_ssh_project:dropbear_ssh:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-PuTTY(?:_Release_)?([\d\.]+)?", re.I), "PuTTY", "cpe:2.3:a:simon_tatham:putty:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bPuTTY\b", re.I), "PuTTY", "cpe:2.3:a:simon_tatham:putty:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-Cisco-\d+\.\d+(?:\.\d+)?", re.I), "Cisco IOS SSH", "cpe:2.3:o:cisco:ios:*:*:*:*:*:*:*:*"),
    (re.compile(r"SSH-2\.0-Cisco", re.I), "Cisco IOS SSH", "cpe:2.3:o:cisco:ios:*:*:*:*:*:*:*:*"),
    # --- FTP ---
    (re.compile(r"ProFTPD\s+(\d+\.\d+(?:\.[\w]+)?)", re.I), "ProFTPD", "cpe:2.3:a:proftpd:proftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"ProFTPD\s+Server", re.I), "ProFTPD", "cpe:2.3:a:proftpd:proftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"\(vsFTPd\s+(\d+\.\d+\.\d+)\)", re.I), "vsftpd", "cpe:2.3:a:vsftpd_project:vsftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"vsFTPd\s+(\d+\.\d+\.\d+)", re.I), "vsftpd", "cpe:2.3:a:vsftpd_project:vsftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bvsFTPd\b", re.I), "vsftpd", "cpe:2.3:a:vsftpd_project:vsftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"Pure-FTPd\s+(\d+\.\d+(?:\.\d+)?)", re.I), "Pure-FTPd", "cpe:2.3:a:pure-ftpd:pure-ftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"Welcome\s+to\s+Pure-FTPd", re.I), "Pure-FTPd", "cpe:2.3:a:pure-ftpd:pure-ftpd:*:*:*:*:*:*:*:*"),
    (re.compile(r"Microsoft\s+FTP\s+Service", re.I), "Microsoft IIS FTP", "cpe:2.3:a:microsoft:internet_information_services:*:*:*:*:*:*:*:*"),
    (re.compile(r"\(MikroTik\s+(\d+\.\d+(?:\.\d+)?)\)", re.I), "MikroTik RouterOS FTP", "cpe:2.3:o:mikrotik:routeros:*:*:*:*:*:*:*:*"),
    (re.compile(r"MikroTik\s+(\d+\.\d+(?:\.\d+)?)", re.I), "MikroTik RouterOS FTP", "cpe:2.3:o:mikrotik:routeros:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bMikroTik\b", re.I), "MikroTik RouterOS FTP", "cpe:2.3:o:mikrotik:routeros:*:*:*:*:*:*:*:*"),
    (re.compile(r"FileZilla\s+Server\s+(\d+\.\d+(?:\.\d+)?)", re.I), "FileZilla Server", "cpe:2.3:a:filezilla-project:filezilla_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"FileZilla\s+Server", re.I), "FileZilla Server", "cpe:2.3:a:filezilla-project:filezilla_server:*:*:*:*:*:*:*:*"),
    (re.compile(r"Serv-U\s+(\d+\.\d+)", re.I), "Serv-U FTP", "cpe:2.3:a:solarwinds:serv-u:*:*:*:*:*:*:*:*"),
    (re.compile(r"\bServ-U\b", re.I), "Serv-U FTP", "cpe:2.3:a:solarwinds:serv-u:*:*:*:*:*:*:*:*"),
    (re.compile(r"WS_FTP\s+Server", re.I), "WS_FTP Server", "cpe:2.3:a:progress:ws_ftp:*:*:*:*:*:*:*:*"),
]


@dataclass(frozen=True)
class ServiceIdentification:
    """Result of banner-based service identification."""
    product: str
    version: str | None  # None if not specified or if captured string is OS/distro name
    cpe: str
    os: str | None = None  # OS/distribution from banner (e.g. Ubuntu), when not a product version


def identify_service(banner: str | None) -> ServiceIdentification | None:
    """
    Identify product and optional version from server banner.
    Returns ServiceIdentification(product, version, cpe, os) or None if no match.
    Version is set only for numeric app versions; distro names (Ubuntu, Debian, ...) go to os.
    CPE uses * for version when not found (full 11-attribute CPE 2.3 string).
    """
    if not banner or not banner.strip():
        return None
    text = banner.replace("\r", " ").strip()
    if not text:
        return None
    # Use first line for matching (banner often multi-line)
    first_line = text.split("\n")[0].strip() if "\n" in text else text
    for pattern, product, cpe_base in _BANNER_PATTERNS:
        m = pattern.search(first_line)
        if m:
            raw = m.group(1).strip() if m.lastindex and m.lastindex >= 1 else None
            # Numeric version -> version; otherwise (e.g. Ubuntu, Debian/GNU) -> os, version stays None
            if raw and _CPE_VERSION_RE.match(raw):
                version, os_str = raw, None
            elif raw:
                version, os_str = None, raw
            else:
                version, os_str = None, None
            cpe_parts = cpe_base.split(":")
            use_in_cpe = (
                bool(version)
                and cpe_base.startswith("cpe:2.3:")
                and len(cpe_parts) >= 6
                and cpe_parts[5] == "*"
            )
            if use_in_cpe:
                injected = list(cpe_parts)
                injected[5] = version
                cpe = ":".join(injected)
            else:
                cpe = cpe_base
            cpe = _cpe_23_normalize(cpe)
            return ServiceIdentification(product=product, version=version, cpe=cpe, os=os_str)
    return None


# SMTP 220: reject lines that name a known MTA/vendor (even if no full banner pattern matched).
_BANNER_MTA_VENDOR_TOKEN: Final[re.Pattern[str]] = re.compile(
    r"\b("
    r"postfix|exim|exchange|sendmail|microsoft|icewarp|zimbra|hmail|haraka|mdaemon|"
    r"mailenable|axigen|courier|qmail|dovecot|kerio|fortimail|cisco|barracuda|mimecast|"
    r"proofpoint|mailgun|sendgrid|mandrill|amazon|ses|gsmtp|zoho|proton|yandex|fastmail|"
    r"outlook|office365|mailstore|trend\s+micro|sophos|power\s*mta|postcow|mailcow|"
    r"communigate|smartermail|hmailserver|oracle\s+communications|asyncos|ironport"
    r")\b",
    re.I,
)

# First line looks like FQDN + ESMTP only (no product fingerprint from identify_service).
_GENERIC_SMTP_BANNER_SHAPE: Final[re.Pattern[str]] = re.compile(
    r"^(\S+\.)+\S+\s+ESMTP(?:\s+(?:ready|server|service(?:\s+ready)?))?\s*$",
    re.I,
)


def is_generic_esmtp_banner(banner: str | None) -> bool:
    """
    True when the SMTP greeting is a minimal hostname + ESMTP style line with no
    identify_service() match and no obvious MTA/vendor token in the first line.
    """
    if not banner or not banner.strip():
        return False
    if identify_service(banner) is not None:
        return False
    first = banner.replace("\r", "").split("\n")[0].strip()
    rest = re.sub(r"^220\s+", "", first, flags=re.I).strip()
    if not rest or _BANNER_MTA_VENDOR_TOKEN.search(rest):
        return False
    if not re.search(r"\bESMTP\b", rest, re.I):
        return False
    return bool(_GENERIC_SMTP_BANNER_SHAPE.match(rest))



from enum import Enum
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *
from typing import List, NamedTuple
from pysnmp.proto.errind import RequestTimedOut
import argparse
from dataclasses import dataclass
from typing import List, Optional

from ptlibs.ptjsonlib import PtJsonLib

from ._base import BaseModule, BaseArgs, Out
from .utils.helpers import text_or_file

    
class VULNS(Enum):
    WeakCommunityName = "PTV-SNMPv2-WEAKCOMMUNITYNAME"
    WeakUsername = "PTV-SNMPv3-WEAKUSERNAME"
    WeakCredentials = "PTV-SNMPv3-WEAKCREDENTIALS"
    Write_2 = "PTV-SNMPv2-WRITEACCESS"
    Write_3 = "PTV-SNMPv3-WRITEACCESS"
    Readmib_3 = "PTV-SNMPv3-READINGMIB"
    Readmib_2 = "PTV-SNMPv2-READINGMIB"

class Credential(NamedTuple):
    username: str | None
    password: str | None


class SNMPVersion(NamedTuple):
    v1: bool | None
    v2c: bool | None
    v3: bool | None

class WriteTestResult(NamedTuple):
    OID: str | None
    creds: str | None   #community for snmpv2
    value: str | None   


class AuthPrivProtocols(NamedTuple):
    auth_protocols: str | None
    priv_protocols: str | None

@dataclass
class SNMPResult:
    version: Optional[SNMPVersion] = None
    communities: Optional[List[str]] = None
    usernames: Optional[List[str]] = None
    credentials: Optional[List[Credential]] = None
    Writetest3: Optional[List[WriteTestResult]] = None
    Writetest2: Optional[List[WriteTestResult]] = None
    Bulk2: Optional[str] = None
    Bulk3: Optional[str] = None

class SNMPArgs(BaseArgs):
    ip: str
    port: int
    command: str
    output: Optional[str] = None
    single_community: str = None
    single_username: str = None
    single_password: str = None
    community_file: str = None
    username_file: str = None
    password_file: str = None
    valid_credentials_file: str = None
    spray: bool = False
    auth_protocols: str = None
    priv_protocols: str = None
    oid: str = "1.3.6"
    oid_format: bool = False
    value: str = "Testvalue123"

    @staticmethod
    def get_help():
        return [
            {"description": ["SNMP Testing Module"]},
            {"usage": ["ptsrvtester snmp <command> <options>"]},
            {"usage_example": [
                "ptsrvtester snmp detection --ip 192.168.1.1",
                "ptsrvtester snmp snmpv2-brute --community-file communities.txt --ip 192.168.1.1",
                "ptsrvtester snmp snmpv3-brute --username-file users.txt --password-file passwords.txt --ip 192.168.1.1"
            ]},
            {"options": [
                ["detection", "<options>", "", "Detect SNMP versions"],
                ["snmpv2-brute", "<options>", "", "SNMPv2 dictionary attack"],
                ["snmpv2-write", "<options>", "", "Test SNMPv2 write permission"],
                ["snmpv2-walk", "<options>", "", "SNMPv2 MIB walk"],
                ["snmpv3-enum", "<options>", "", "SNMPv3 user enumeration"],
                ["snmpv3-brute", "<options>", "", "SNMPv3 credentials bruteforce"],
                ["snmpv3-walk", "<options>", "", "SNMPv3 MIB walk"],
                ["snmpv3-write", "<options>", "", "Test SNMPv3 write permissions"],
                ["", "", "", ""],
                ["-h", "--help", "", "Show this help message and exit"],
            ]}
        ]

    def add_subparser(self, name: str, subparsers) -> None:
        """Adds a subparser of SNMP arguments"""

        examples = """example usage:
    ptsrvtester snmp detection --ip 192.168.1.1 --port 161
    ptsrvtester snmp snmpv2-brute --community-file communities.txt --ip 192.168.1.1 --port 161
    ptsrvtester snmp snmpv3-brute --username-file users.txt --password-file passwords.txt --ip 192.168.1.1 --port 161"""

        parser = subparsers.add_parser(
            name,
            epilog=examples,
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        if not isinstance(parser, argparse.ArgumentParser):
            raise TypeError

        snmp_subparsers = parser.add_subparsers(dest="command", help="Select SNMP command", required=True)

        # SNMP Version Detection
        detection = snmp_subparsers.add_parser("detection", help="Detect SNMP versions")
        detection.add_argument("-ip", "--ip", required=True, help="Target IP address")
        detection.add_argument("-p", "--port", type=int,default = 161, help="Target port")

        # SNMPv2 Brute Force
        snmpv2_brute_parser = snmp_subparsers.add_parser("snmpv2-brute", help="SNMPv2 dictionary attack")
        snmpv2_brute_parser.add_argument("-ip", "--ip", required=True, help="Target IP address")
        snmpv2_brute_parser.add_argument("-p", "--port", type=int, default = 161, help="Target port")
        snmpv2_brute_parser.add_argument("-o", "--output",  help="File to save the output results.")

        user_group1 = snmpv2_brute_parser.add_mutually_exclusive_group(required=True)
        user_group1.add_argument("-c", "--single-community", "--community", help="Single community string")
        user_group1.add_argument("-cf", "--community-file", help="File containing community strings")

        # SNMPv2 Write Permission
        snmpv2_write_parser = snmp_subparsers.add_parser("snmpv2-write", help="Test SNMPv2 write permission")
        snmpv2_write_parser.add_argument("-ip", "--ip", required=True, help="Target IP address")
        snmpv2_write_parser.add_argument("-p", "--port", type=int, default = 161, help="Target port")
        snmpv2_write_parser.add_argument("-v", "--value", default="Testvalue123", help="Value to write to the specified OID (default: 'Testvalue123')")

        user_group2 = snmpv2_write_parser.add_mutually_exclusive_group(required=True)
        user_group2.add_argument("-c", "--single-community", "--community", help="Single community string")
        user_group2.add_argument("-cf", "--community-file", help="File containing community strings")

        # SNMPv2 GetBulk (Walk)
        snmpv2_getbulk_parser = snmp_subparsers.add_parser("snmpv2-walk", help="SNMPv2 MIB walk")
        snmpv2_getbulk_parser.add_argument("-ip","--ip", required=True, help="Target IP address")
        snmpv2_getbulk_parser.add_argument("-p","--port", type=int, default = 161, help="Target port")
        snmpv2_getbulk_parser.add_argument("-oid","--oid", default="1.3.6", help="OID to start from")
        snmpv2_getbulk_parser.add_argument("-of","--oid-format", action="store_true", help="Use human readable OID format")
        snmpv2_getbulk_parser.add_argument("-o","--output", help="File to save the output results.")

        user_group3 = snmpv2_getbulk_parser.add_mutually_exclusive_group(required=True)
        user_group3.add_argument("-c", "--single-community", "--community", help="Single community string")
        user_group3.add_argument("-cf", "--community-file", help="File containing community strings")

        # SNMPv3 User Enumeration
        user_enum_parser = snmp_subparsers.add_parser("snmpv3-enum", help="SNMPv3 user enumeration")
        user_enum_parser.add_argument("-ip","--ip", required=True, help="Target IP address")
        user_enum_parser.add_argument("-p","--port", type=int, default = 161, help="Target port")
        user_enum_parser.add_argument("-o", "--output", help="File to save the output results.")

        user_group4 = user_enum_parser.add_mutually_exclusive_group(required=True)
        user_group4.add_argument("-u", "--single-username", help="Single username")
        user_group4.add_argument("-ul", "--username-file", help="File containing usernames")

        # SNMPv3 Brute Force
        snmpv3_brute_parser = snmp_subparsers.add_parser("snmpv3-brute", help="SNMPv3 credentials bruteforce")
        snmpv3_brute_parser.add_argument("-ip", "--ip", required=True, help="Target IP address")
        snmpv3_brute_parser.add_argument("-p", "--port", type=int, default = 161, help="Target port")
        snmpv3_brute_parser.add_argument("-ap", "--auth-protocols", help="Authentication protocol")
        snmpv3_brute_parser.add_argument("-pp", "--priv-protocols", help="Private protocol")
        snmpv3_brute_parser.add_argument("-o", "--output", help="File to save the output results.")
        snmpv3_brute_parser.add_argument("-s", "--spray", action="store_true", help="Enable spray mode")
  
        user_group6 = snmpv3_brute_parser.add_mutually_exclusive_group(required=True)
        user_group6.add_argument("-u", "--single-username", help="Single username")
        user_group6.add_argument("-ul", "--username-file", help="File containing usernames")

        user_group7 = snmpv3_brute_parser.add_mutually_exclusive_group(required=True)
        user_group7.add_argument("-pw", "--single-password", help="Single password")
        user_group7.add_argument("-pl", "--password-file", help="File containing passwords")


        # SNMPv3 GetBulk (Walk)
        snmpv3_getbulk_parser = snmp_subparsers.add_parser("snmpv3-walk", help="SNMPv3 MIB walk")
        snmpv3_getbulk_parser.add_argument("-ip", "--ip", required=True, help="Target IP address")
        snmpv3_getbulk_parser.add_argument("-p", "--port", type=int, default = 161, help="Target port")
        snmpv3_getbulk_parser.add_argument("-u", "--single-username", help="Single username")
        snmpv3_getbulk_parser.add_argument("-pw", "--single-password", help="Single password")
        snmpv3_getbulk_parser.add_argument("-ap", "--auth-protocols", help="Authentication protocol")
        snmpv3_getbulk_parser.add_argument("-pp", "--priv-protocols", help="Private protocol")
        snmpv3_getbulk_parser.add_argument("-oid", "--oid", default="1.3.6", help="OID to start from")
        snmpv3_getbulk_parser.add_argument("-of", "--oid-format", action="store_true", help="Use human readable OID format")
        snmpv3_getbulk_parser.add_argument("-o", "--output", help="File to save the output results.")

        # SNMPv3 Write Permission
        snmpv3_write = snmp_subparsers.add_parser("snmpv3-write", help="Test SNMPv3 write permissions")
        snmpv3_write.add_argument("-ip", "--ip", required=True, help="Target IP address")
        snmpv3_write.add_argument("-p", "--port", type=int, default = 161, help="Target port")
        snmpv3_write.add_argument("-u", "--single-username", help="Single username")
        snmpv3_write.add_argument("-pw", "--single-password", help="Single password")
        snmpv3_write.add_argument("-cred", "--valid-credentials-file", help="File containing valid credentials")
        snmpv3_write.add_argument("-ap", "--auth-protocols", help="Authentication protocol")
        snmpv3_write.add_argument("-pp", "--priv-protocols", help="Private protocol")
        snmpv3_write.add_argument("-v", "--value", default="Testvalue123", help="Value to write to the specified OID (default: 'Testvalue123')")


class SNMP(BaseModule):
    @staticmethod
    def module_args():
        return SNMPArgs()

    def __init__(self, args: BaseArgs, ptjsonlib: PtJsonLib):
        self.args = args  # type: SNMPArgs
        self.ptjsonlib = ptjsonlib
        self.results: SNMPResult | None = None

    def run(self) -> None:
        """Main SNMP execution logic"""

        self.results = SNMPResult()

        if self.args.command == "detection":
            self.results.version = asyncio.run(self.version_detection())

        elif self.args.command == "snmpv2-brute":
            self.results.communities = asyncio.run(self.snmpv2_brute())

        elif self.args.command == "snmpv3-brute":
            self.results.credentials = asyncio.run(self.snmpv3_brute())

        elif self.args.command == "snmpv3-enum":
            self.results.usernames = asyncio.run(self.user_enum())

        elif self.args.command == "snmpv2-write":
            self.results.Writetest2 = asyncio.run(self.test_snmpv2_write_permission())

        elif self.args.command == "snmpv3-write":
            self.results.Writetest3 = asyncio.run(self.test_snmpv3_write_permissions())

        elif self.args.command == "snmpv2-walk":
            self.results.Bulk2 = asyncio.run(self.getBulk_SNMPv2())

        elif self.args.command == "snmpv3-walk":
            self.results.Bulk3 = asyncio.run(self.getBulk_SNMPv3())
        
        else:
            self.ptprint("Unknown command for SNMP module.", out=Out.WARNING)

     # Map protocol OIDs to human-readable names
    PROTOCOL_NAMES = {
        usmHMACMD5AuthProtocol: "usmHMACMD5AuthProtocol",
        usmHMACSHAAuthProtocol: "usmHMACSHAAuthProtocol",
        usmHMAC128SHA224AuthProtocol: "usmHMAC128SHA224AuthProtocol",
        usmHMAC192SHA256AuthProtocol: "usmHMAC192SHA256AuthProtocol",
        usmHMAC256SHA384AuthProtocol: "usmHMAC256SHA384AuthProtocol",
        usmHMAC384SHA512AuthProtocol: "usmHMAC384SHA512AuthProtocol",
        usmDESPrivProtocol: "usmDESPrivProtocol",
        usmAesCfb128Protocol: "usmAesCfb128Protocol",
        usmAesCfb192Protocol: "usmAesCfb192Protocol",
        usmAesCfb256Protocol: "usmAesCfb256Protocol",
        None: "None",
    }

    def drawDoubleLine(self):
        self.ptprint ('=' * 75)

    def write_to_file(self, message_or_messages: str | list[str]):
        """
            File Output.
        """
        try:
            with open(self.args.output, "a") as f:
                if isinstance(message_or_messages, str):
                    f.write(message_or_messages + "\n")
                elif isinstance(message_or_messages, list):
                    for message in message_or_messages:
                        f.write(message + "\n")
        except FileNotFoundError:
            raise argparse.ArgumentError(None, f"File not found: '{self.args.output}'")
        except PermissionError:
            raise argparse.ArgumentError(
                None, f"Cannot write file (permission denied): '{self.args.output}'"
            )
        except OSError as e:
            raise argparse.ArgumentError(None, f"Cannot write file '{self.args.output}': {e}")

    def _text_or_file(self, text: str | None, file_path: str | None) -> List[str]:

        values = text_or_file(text.strip() if text else None, file_path)
        return [v.strip() for v in values if v.strip()]

    # Function for getBulk SNMPv2/SNMPv3
    def format_timeticks(self, value):
        """
            Convert Timeticks to a human-readable string.
        """
        ticks = int(value)
        days, remainder = divmod(ticks, 8640000)  # 1 day = 8640000 timeticks
        hours, remainder = divmod(remainder, 360000)
        minutes, remainder = divmod(remainder, 6000)
        seconds = remainder // 100
        return f"{days} day, {hours}:{minutes:02}:{seconds:02}.{remainder % 100}"

    async def version_detection(self) -> SNMPVersion | None:
        """
           Detects the SNMP version supported by the target device.

           Parameters:
           - self.ip (str): The IP address of the target device.
           - self.port (int): The port number for SNMP communication.

           Returns:
           - SNMPVersion: An object containing three boolean attributes (`v1`, `v2c`, `v3`), each indicating
             whether the corresponding SNMP version is supported by the target device.
        """

        # Struct data
        v1: bool = False
        v2c: bool = False
        v3: bool = False

        ###########################################################################################
        # Detect v1                                                                               #
        ###########################################################################################
        iterator = await get_cmd(
            SnmpEngine(),
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create((self.args.ip, self.args.port)),
            ContextData(),
            ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
        )

        errorIndication, errorStatus, errorIndex, varBinds = iterator

        if errorIndication:
            self.ptprint(f"Error!: {errorIndication}", out=Out.ERROR)
        elif errorStatus:
            self.ptprint(
                "{} at {}".format(
                    errorStatus.prettyPrint(),
                    errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
                ),
                out=Out.ERROR
            )

        else:
            v1 = True
            for varBind in varBinds:
                self.ptprint("Success!: ", end="", out=Out.OK)
                self.ptprint(" = ".join([x.prettyPrint() for x in varBind]))

        ###########################################################################################
        # Detect v2c                                                                              #
        ###########################################################################################
        iterator = await get_cmd(
            SnmpEngine(),
            CommunityData("public", mpModel=1),
            await UdpTransportTarget.create((self.args.ip, self.args.port)),
            ContextData(),
            ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
        )

        errorIndication, errorStatus, errorIndex, varBinds = iterator

        if errorIndication:
             self.ptprint(f"Error!: {errorIndication}", out=Out.ERROR)

        elif errorStatus:
            self.ptprint("Error!:", out=Out.ERROR)
            self.ptprint(
                "{} at {}".format(
                    errorStatus.prettyPrint(),
                    errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
                ),
                out=Out.ERROR
            )

        else:
            v2c = True
            for varBind in varBinds:
                self.ptprint("Success!: ", end="", out=Out.OK)  # můžeš nastavit `out=Out.SUCCESS` pokud máš definováno
                self.ptprint(" = ".join([x.prettyPrint() for x in varBind]))

        ###########################################################################################
        # Detect v3                                                                               #
        ###########################################################################################
        iterator = await get_cmd(
            SnmpEngine(),
            UsmUserData("pentest"),
            await UdpTransportTarget.create((self.args.ip, self.args.port)),
            ContextData(),
        )

        errorIndication, errorStatus, errorIndex, varBinds = iterator

        if errorIndication:
            if isinstance(errorIndication, RequestTimedOut):
                self.ptprint(f"Error!: {errorIndication}", out=Out.ERROR)
            else:
                self.ptprint(f"Success!: {errorIndication}", out=Out.OK)
                v3 = True

        elif errorStatus:
            self.ptprint(
                "{} at {}".format(
                    errorStatus.prettyPrint(),
                    errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
                ),
                out=Out.ERROR
            )

        else:
            v3 = True
            for varBind in varBinds:
                self.ptprint(" = ".join([x.prettyPrint() for x in varBind]))

        self.ptprint(str(SNMPVersion(v1, v2c, v3)))
        return SNMPVersion(v1, v2c, v3)

    async def snmpv2_brute(self) -> List[str]:

        """
           Performs a dictionary attack on SNMPv2/1 to find valid communities.

           Parameters:
           - self.single_community (str): A single community string for SNMPv2/1 authentication.
           - self.community_file (str): Path to a file containing a list of communities for the dictionary attack.
           - self.ip (str): The IP address of the target device.
           - self.port (int): The port number for SNMP communication.
           - self.output (bool): If True, writes valid credentials to a file.

           Returns:
           - list[Credential]: A list of valid communities found during the attack.
           - None: If no credentials are found or required inputs are missing.
        """

        if not self.args.community_file and not self.args.single_community:
            self.ptprint("Error: Neither a community file nor a single community string was provided.", out=Out.WARNING)
            return []
        self.drawDoubleLine()
        self.ptprint("Starting a dictionary attack on SNMPv2...", title=True)
        self.drawDoubleLine()
        communities = self._text_or_file(self.args.single_community, self.args.community_file)
        valid_communities = []

        for community in communities:
            iterator = get_cmd(SnmpEngine(),
                               CommunityData(community),
                               await UdpTransportTarget.create((self.args.ip, self.args.port), timeout=0.1),
                               # Initialize transport target correctly
                               ContextData(),
                               ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)))
            errorIndication, errorStatus, errorIndex, varBinds = await iterator

            if not errorIndication and not errorStatus:
                self.ptprint(f"Valid community string found: {community}", out=Out.OK)
                valid_communities.append(community)
            else:
                self.ptprint(f"Error: {errorIndication or errorStatus} for {community}", out=Out.ERROR)

        if valid_communities:
            self.ptprint("\n")
            self.ptprint("Valid communities:", out=Out.INFO)
            for community in valid_communities:
                self.ptprint(community)
            if self.args.output:
                for community in valid_communities:
                    self.write_to_file(community)

        else:
            self.ptprint("\nNo valid communities found :(", out=Out.INFO)
        return valid_communities

    async def user_enum(self) -> list[str]:


        # Users from input
        users: list[str] = self._text_or_file(self.args.single_username, self.args.username_file)

        self.drawDoubleLine()
        self.ptprint("Starting username enumeration...", title=True)
        self.drawDoubleLine()
        valid_usernames = set()

        for username in users:
            try:
                iterator = get_cmd(
                    SnmpEngine(),
                    UsmUserData(username, "userenumeration", authProtocol=None, privProtocol=None),
                    await UdpTransportTarget.create((self.args.ip, self.args.port)),
                    ContextData(),
                    ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                )
                errorIndication, errorStatus, errorIndex, varBinds = await iterator

                if not errorIndication and not errorStatus:
                    self.ptprint(f"Valid username found: {username}", out=Out.OK)
                    valid_usernames.add(username)
                elif "Wrong SNMP PDU digest" in str(errorIndication):
                    self.ptprint(f"Potential valid username: {username}", out=Out.OK)
                    valid_usernames.add(username)
                else:
                    
                    self.ptprint(f"Error for username {username}: {errorIndication or errorStatus}", out=Out.ERROR)

            except Exception as e:
                self.ptprint(f"Error for username {username}: {e}", out=Out.ERROR)

        if valid_usernames:
            self.ptprint("\n")
            self.ptprint("Potential valid usernames:", out=Out.INFO)
            for username in valid_usernames:
                self.ptprint(username)
                if self.args.output:
                    self.write_to_file(username)
        else:
            self.ptprint("No valid usernames found.", out=Out.INFO)

        return list(valid_usernames)

    async def snmpv3_brute(self) -> list[Credential] | None:

        """
            Performs a dictionary attack on SNMPv3 to find valid credentials.

            Parameters:
            - self.single_username (str): A single username for SNMPv3 authentication.
            - self.single_password (str): A single password for SNMPv3 authentication.
            - self.username_file (str): Path to a file containing a list of usernames for the dictionary attack.
            - self.password_file (str): Path to a file containing a list of passwords for the dictionary attack.
            - self.auth_protocols (obj): The authentication protocol to use (e.g., usmHMACSHAAuthProtocol). Defaults to a set of standard protocols if not provided.
            - self.priv_protocols (obj): The encryption protocol to use (e.g., usmDESPrivProtocol). Defaults to a set of standard protocols if not provided.
            - self.spray (bool): Determines whether to try all passwords for each username (False) or all usernames for each password (True).
            - self.ip (str): The IP address of the target device.
            - self.port (int): The port number for SNMP communication.
            - self.output (bool): If True, writes valid credentials to a file.

            Returns:
            - list[Credential]: A list of valid credentials (username and password pairs) found during the attack.
            - None: If no credentials are found or required inputs are missing.
        """

        # Warning
        if not self.args.username_file and not self.args.single_username:
            self.ptprint("Error: Neither a username file nor a single username was provided.", out=Out.WARNING)
            return None

        # Warning
        if not self.args.password_file and not self.args.single_password:
            self.ptprint("Error: Neither a password file nor a single password was provided.", out=Out.WARNING)
            return None

        # Users and passwords from input
        users = self._text_or_file(self.args.single_username, self.args.username_file)
        passwords = self._text_or_file(self.args.single_password, self.args.password_file)
        valid_usernames = set()

        # setting the hash function for bruteforce
        default_auth_protocols = [
            usmHMACSHAAuthProtocol,
            usmHMACMD5AuthProtocol,
            usmHMAC128SHA224AuthProtocol,
            usmHMAC192SHA256AuthProtocol,
            usmHMAC256SHA384AuthProtocol,
            usmHMAC384SHA512AuthProtocol
        ]
        # setting the encryption function for bruteforce
        default_priv_protocols = [
            usmDESPrivProtocol,
            usmAesCfb128Protocol,
            usmAesCfb192Protocol,
            usmAesCfb256Protocol
        ]

        # If protocols are not set, perform username enumeration first
        if (self.args.auth_protocols is None or self.args.priv_protocols is None) and self.args.username_file:
            self.ptprint("No auth or priv protocols set." , out=Out.TITLE)
            users = await self.user_enum()
            valid_usernames = set(users)
            if not users:
                self.ptprint("\n")
                self.ptprint("Sorry, it is not possible to find valid credentials with these usernames", out=Out.ERROR)
                return None
            
        PROTOCOL_OBJECTS = {v: k for k, v in self.PROTOCOL_NAMES.items()}

        if isinstance(self.args.auth_protocols, str):
            self.args.auth_protocols = PROTOCOL_OBJECTS.get(self.args.auth_protocols, None)
            if self.args.auth_protocols is None:
                self.ptprint("Warning: Unknown authentication protocol string. Using defaults.", out=Out.INFO)


        if isinstance(self.args.priv_protocols, str):
            self.args.priv_protocols = PROTOCOL_OBJECTS.get(self.args.priv_protocols, None)
            if self.args.priv_protocols is None:
                self.ptprint("Warning: Unknown privacy protocol string. Using defaults.", out=Out.INFO)

        auth_protocols = [self.args.auth_protocols] if self.args.auth_protocols else default_auth_protocols
        priv_protocols = [self.args.priv_protocols] if self.args.priv_protocols else default_priv_protocols

        protocols = [AuthPrivProtocols(a, p) for a in auth_protocols for p in priv_protocols]

        # Spray logic
        if self.args.spray:
            creds = [Credential(u, p) for p in passwords for u in users]
        else:
            creds = [Credential(u, p) for u in users for p in passwords]

        found_credentials = []  # store valid found credentials
        successful_protocol = None  # Track the successful protocol combination
        valid_usernames = set()

        # starting the attack
        self.ptprint("\n")
        self.drawDoubleLine()
        self.ptprint("Starting a dictionary attack on SNMPv3...", title=True)
        self.drawDoubleLine()

        for protocol in protocols:
            if successful_protocol:
                # If a valid protocol was found, skip other combinations
                if protocol != successful_protocol:
                    continue
            for cred in creds:
                try:
                    iterator = get_cmd(SnmpEngine(),
                                       UsmUserData(cred.username, cred.password, authProtocol=protocol.auth_protocols, privProtocol=protocol.priv_protocols),
                                       await UdpTransportTarget.create((self.args.ip, self.args.port)),
                                       ContextData(),
                                       ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)))
                    errorIndication, errorStatus, errorIndex, varBinds = await iterator

                    if not errorIndication and not errorStatus:
                        found_credentials.append(cred)
                        successful_protocol = protocol
                        valid_usernames.add(cred.username)
                        auth_name = self.PROTOCOL_NAMES.get(successful_protocol.auth_protocols, "Unknown Protocol")
                        priv_name = self.PROTOCOL_NAMES.get(successful_protocol.priv_protocols, "Unknown Protocol")
                        self.ptprint(f"Valid credentials found: Username: {cred.username}, Password: {cred.password}", out=Out.OK)
                        self.ptprint(f"Successful Authentication and Private protocols are: {auth_name} and {priv_name}", out=Out.INFO)
                    elif "Wrong SNMP PDU digest" in str(errorIndication):
                        self.ptprint(f"Digest match (likely valid username - Try different password or protocols): {cred.username}", out=Out.INFO)
                        valid_usernames.add(cred.username)
                    elif "Unknown USM user" in str(errorIndication):
                        self.ptprint(f"Error: Unknown user: {cred.username}", out=Out.ERROR)
                    else:
                        self.ptprint(f"Error: {errorIndication or errorStatus} for {cred.username}/{cred.password}", out=Out.ERROR)

                except Exception as e:
                    self.ptprint(f"Error: {cred.username}/{cred.password}: {e}", out=Out.ERROR)

        if valid_usernames:
            self.ptprint("\n")
            self.ptprint("Potential valid usernames:", out=Out.INFO)
            for username in valid_usernames:
                self.ptprint(username)

        if self.args.output and found_credentials:
            results = [f"Username: {cred.username}, Password: {cred.password}" for cred in found_credentials]
            self.write_to_file(results)

        if found_credentials:
            self.ptprint("\n")
            self.ptprint("Found credentials:", out=Out.INFO)
            for cred in found_credentials:
                self.ptprint(f"Username: {cred.username}, Password: {cred.password}")

        if successful_protocol:
            auth_name = self.PROTOCOL_NAMES.get(successful_protocol.auth_protocols, "Unknown Protocol")
            priv_name = self.PROTOCOL_NAMES.get(successful_protocol.priv_protocols, "Unknown Protocol")
            self.ptprint("\n")
            self.ptprint(f"Successful Authentication and Private protocols are: {auth_name} and {priv_name}", out=Out.INFO)

        else:
            self.ptprint("\n")
            self.ptprint("No valid credentials found :(", out=Out.ERROR)

        return found_credentials

    async def test_snmpv3_write_permissions(self) -> list[WriteTestResult]:
        """
            Tests SNMPv3 write permissions by attempting to set a value on the target device.

            Parameters:
            - self.single_username (str): A single username for SNMPv3 authentication.
            - self.single_password (str): A single password for SNMPv3 authentication.
            - self.auth_protocols (obj): The authentication protocol (e.g., usmHMACSHAAuthProtocol). Defaults to usmHMACSHAAuthProtocol if not provided.
            - self.priv_protocols (obj): The encryption protocol (e.g., usmDESPrivProtocol). Defaults to usmDESPrivProtocol if not provided.
            - self.valid_credentials_file (str): Path to a file containing multiple valid credentials in the format `username: value, password: value`.
            - self.ip (str): The IP address of the target device.
            - self.port (int): The port number.

            Returns:
            - None: Prints the results of the write test, including success or failure messages.
        """
        results: list[WriteTestResult] = []
        default_auth_protocol = usmHMACSHAAuthProtocol
        default_priv_protocol = usmDESPrivProtocol

        PROTOCOL_OBJECTS = {v: k for k, v in self.PROTOCOL_NAMES.items()}

        if isinstance(self.args.auth_protocols, str):
            self.args.auth_protocols = PROTOCOL_OBJECTS.get(self.args.auth_protocols, None)
            if self.args.auth_protocols is None:
                self.ptprint("Warning: Unknown authentication protocol string. Using defaults.", out=Out.INFO)


        if isinstance(self.args.priv_protocols, str):
            self.args.priv_protocols = PROTOCOL_OBJECTS.get(self.args.priv_protocols, None)
            if self.args.priv_protocols is None:
                self.ptprint("Warning: Unknown privacy protocol string. Using defaults.", out=Out.INFO)

        if not self.args.auth_protocols:
            self.ptprint("Be aware that authentication protocol was not provided, so it is set as usmHMACSHAAuthProtocol", out=Out.INFO)
            self.args.auth_protocols = default_auth_protocol

        if not self.args.priv_protocols:
            self.ptprint("Be aware that private protocol was not provided, so it is set as usmDESPrivProtocol", out=Out.INFO)
            self.args.priv_protocols = default_priv_protocol

        creds = []

        Protocols = AuthPrivProtocols(self.args.auth_protocols, self.args.priv_protocols)

        if self.args.single_username and self.args.single_password:
            creds.append(Credential(self.args.single_username, self.args.single_password))
        elif self.args.valid_credentials_file:
            inputs = self._text_or_file(None, self.args.valid_credentials_file)
            for line in inputs:
                # Parse username and password directly from the line
                parts = line.split(", ")
                if len(parts) == 2:
                    try:
                        username = parts[0].split(": ")[1]
                        password = parts[1].split(": ")[1]
                        creds.append(Credential(username, password))
                    except IndexError:
                        self.ptprint(f"Invalid line format: {line}", out=Out.WARNING)
                else:
                    self.ptprint(f"Invalid format: {line}", out=Out.WARNING)
        else:
            self.ptprint("Error: Provide either single username/password or a file with credentials.", out=Out.WARNING)
            return
        self.drawDoubleLine()
        self.ptprint("Starting SNMPv3 write permission test...", title=True)
        self.drawDoubleLine()
        

        for cred in creds:
            try:
                self.ptprint(f"Testing write permission for user: {cred.username} with password: {cred.password}", out=Out.INFO)
                iterator = set_cmd(
                    SnmpEngine(),
                    UsmUserData(cred.username, cred.password, authProtocol=Protocols.auth_protocols, privProtocol=Protocols.priv_protocols),
                    await UdpTransportTarget.create((self.args.ip, self.args.port)),
                    ContextData(),
                    ObjectType(ObjectIdentity("SNMPv2-MIB", "sysName", 0), OctetString(self.args.value))
                )

                errorIndication, errorStatus, errorIndex, varBinds = await iterator

                if not errorIndication and not errorStatus:
                    self.ptprint("Test was successful!", out=Out.OK)
                    for varBind in varBinds:
                        self.ptprint(f"OID: {varBind[0]} was set to {varBind[1]}")
                        self.ptprint(f"Note: Attribute was modified for testing purposes. Don't forget to revert it back if necessary.", out=Out.INFO)
                        results.append(WriteTestResult(
                        OID=str(varBind[0]),
                        creds=f"{cred.username or 'None'}:{cred.password or 'None'}",
                        value=str(varBind[1])
                        ))
                else:
                    self.ptprint(f"Test failed: {errorIndication or errorStatus}", out=Out.ERROR)

            except Exception as e:
                self.ptprint(f"Exception occurred: {e}", out=Out.WARNING)

        return results
    
    async def test_snmpv2_write_permission(self) -> list[WriteTestResult]:
        """
            Tests SNMPv2 write permissions by attempting to set a value on the target device.

            Parameters:
            - self.single_community (str): A single community string for SNMPv2/1 authentication.
            - self.community_file (str): Path to a file containing multiple valid community strings.
            - self.ip (str): The IP address of the target device.
            - self.port (int): The port number.

            Returns:
            - None: Prints the results of the write test, including success or failure messages.
        """
        results: list[WriteTestResult] = []
        if not self.args.community_file and not self.args.single_community:
            self.ptprint("Error: Neither a community file nor a single community string was provided.", out=Out.WARNING)
            return results

        communities = self._text_or_file(self.args.single_community, self.args.community_file)
        self.drawDoubleLine()
        self.ptprint("Starting SNMPv2 write permission test...", title=True)
        self.drawDoubleLine()

        for community in communities:
            try:
                self.ptprint(f"Testing write permission for community string: {community}", out=Out.INFO)
                iterator = set_cmd(
                    SnmpEngine(),
                    CommunityData(community),
                    await UdpTransportTarget.create((self.args.ip, self.args.port)),
                    ContextData(),
                    ObjectType(ObjectIdentity("SNMPv2-MIB", "sysName", 0), OctetString(self.args.value))
                )

                errorIndication, errorStatus, errorIndex, varBinds = await iterator

                if not errorIndication and not errorStatus:
                    self.ptprint("Write was successful!", out=Out.OK)
                    for varBind in varBinds:
                        self.ptprint(f"OID: {varBind[0]} was set to {varBind[1]}")
                        self.ptprint(f"Note: Attribute was modified for testing purposes. Don't forget to revert it back if necessary.", out=Out.INFO)
                        results.append(WriteTestResult(
                        OID=str(varBind[0]),
                        creds=f"{community}",
                        value=str(varBind[1])
                        ))
                else:
                    self.ptprint(f"Write failed: {errorIndication or errorStatus}", out=Out.ERROR)

            except Exception as e:
                self.ptprint(f"Exception occurred: {e}", out=Out.WARNING)

        return results

    async def getBulk_SNMPv2(self) -> str:

        """
           Executes an SNMPv2 bulk walk on the target device to retrieve MIB object values based on the specified OID.

           Parameters:
           - self.single_community (str): The community string for SNMPv2 authentication.
           - self.oid (str): The starting OID. Default is "1.3.6" if not provided.
           - self.oid_format (bool): Determines if the OID should be converted to a humanreadable format.
           - self.output (bool): Indicates whether the results should be saved to a file.
           - self.ip (str): The IP address of the target device.
           - self.port (int): The port number.

           Returns:
           - results (list): A list of formatted strings containing OID-value pairs retrieved from the target device.
       """

        if not self.args.community_file and not self.args.single_community:
            self.ptprint("Neither a community file nor a single community string was provided. Defaulting to 'public'.", out=Out.WARNING)
            self.args.single_community = "public"

        communities = self._text_or_file(self.args.single_community, self.args.community_file)

        self.drawDoubleLine()
        self.ptprint("Starting SNMPv2 bulk walk...", title=True)
        self.drawDoubleLine()
        results = []
        # for json
        result = None

        for community in communities:
            self.ptprint(f"Trying community: {community}", out=Out.INFO)
            try:
                # Use walk_cmd to traverse the MIB
                objects = walk_cmd(
                    SnmpEngine(),
                    CommunityData(community),
                    await UdpTransportTarget.create((self.args.ip, self.args.port)),
                    ContextData(),
                    ObjectType(ObjectIdentity(self.args.oid))
                )

                # Iterate over the returned OID-value pairs
                async for errorIndication, errorStatus, errorIndex, varBinds in objects:
                    if errorIndication:
                        self.ptprint(f"Error: {errorIndication}", out=Out.ERROR)
                        break
                    elif errorStatus:
                        self.ptprint(f"Error: {errorStatus.prettyPrint()} at {errorIndex}", out=Out.ERROR)
                        break
                    else:
                        for oid, value in varBinds:
                            if self.args.oid_format:
                                oid = oid.prettyPrint()  # Convert OID to string
                            value_type = value.__class__.__name__.upper()  # Get the value type
                            value_str = value.prettyPrint()  # Convert value to string

                            # Format the value type and content
                            if value_type == "OCTET STRING":
                                value_output = f'STRING: "{value_str}"'
                            elif value_type == "OBJECT IDENTIFIER":
                                value_output = f'OID: {value}'
                            elif value_type == "TIMETICKS":
                                value_output = f'Timeticks: ({value_str}) {self.format_timeticks(value)}'
                            elif value_type == "INTEGER":
                                value_output = f'INTEGER: {value_str}'
                            else:
                                value_output = value_str  # Default for other types

                            # Construct the final formatted string
                            formatted_output = f"{oid} = {value_output}"
                            self.ptprint(formatted_output)
                            results.append(formatted_output)

                # Stop the loop if results are found
                if results:
                    self.ptprint(f"Results found with community '{community}', stopping further attempts.", out=Out.OK)
                    result = "success"
                    break
                    

            except Exception as e:
                self.ptprint(f"Exception occurred for community '{community}': {e}", out=Out.WARNING)
                continue  # Move to the next community in case of errors
        if self.args.output:
            self.write_to_file(results)
        return result

    async def getBulk_SNMPv3(self) -> str:
        """
            Executes an SNMPv3 bulk walk on the target device to retrieve MIB object values based on the specified OID.

            Parameters:
            - self.single_username (str): The username for SNMPv3 authentication.
            - self.single_password (str): The password for SNMPv3 authentication.
            - self.auth_protocols (obj): The authentication protocol (e.g., usmHMACSHAAuthProtocol).
            - self.priv_protocols (obj): The encryption protocol (e.g., usmDESPrivProtocol).
            - self.oid (str): The starting OID. Default is "1.3.6" if not provided.
            - self.oid_format (bool): Determines if the OID should be converted to a humanreadable format.
            - self.output (bool): Indicates whether the results should be saved to a file.
            - self.ip (str): The IP address of the target device.
            - self.port (int): The port number.

            Returns:
            - results (list): A list of formatted strings containing OID-value pairs retrieved from the target device.
        """
        #maps user defined string to oid format of protocol
        PROTOCOL_OBJECTS = {v: k for k, v in self.PROTOCOL_NAMES.items()}

        if not self.args.single_username:
            self.ptprint("\nUsername was not provided, Set the username to Start the SNMPv3 walk", out=Out.WARNING)
            return []

        if not self.args.single_password:
            self.ptprint("\nPassword was not provided, Set the password to Start the SNMPv3 walk", out=Out.WARNING)
            return []

        PROTOCOL_OBJECTS = {v: k for k, v in self.PROTOCOL_NAMES.items()}

        if isinstance(self.args.auth_protocols, str):
            self.args.auth_protocols = PROTOCOL_OBJECTS.get(self.args.auth_protocols, None)
            if self.args.auth_protocols is None:
                self.ptprint("Warning: Unknown authentication protocol string. Using defaults.", out=Out.INFO)


        if isinstance(self.args.priv_protocols, str):
            self.args.priv_protocols = PROTOCOL_OBJECTS.get(self.args.priv_protocols, None)
            if self.args.priv_protocols is None:
                self.ptprint("Warning: Unknown privacy protocol string. Using defaults.", out=Out.INFO)

        if not self.args.auth_protocols:
            self.ptprint("Be aware that authentication protocol was not provided, so it is set as usmHMACSHAAuthProtocol", out=Out.INFO)
            self.args.auth_protocols = usmHMACSHAAuthProtocol

        if not self.args.priv_protocols:
            self.ptprint("Be aware that private protocol was not provided, so it is set as usmAesCfb128Protocol", out=Out.INFO)
            self.args.priv_protocols = usmAesCfb128Protocol

        Protocols = AuthPrivProtocols(self.args.auth_protocols, self.args.priv_protocols)

        if self.args.oid is None:
            self.args.oid = "1.3.6"

        self.drawDoubleLine()
        self.ptprint("Starting SNMPv3 bulk walk...", title=True)
        self.drawDoubleLine()
        results = None

        objects = walk_cmd(
            SnmpEngine(),
            UsmUserData(self.args.single_username, self.args.single_password, authProtocol=Protocols.auth_protocols, privProtocol=Protocols.priv_protocols),
            await UdpTransportTarget.create((self.args.ip, self.args.port)),
            ContextData(),
            ObjectType(ObjectIdentity(self.args.oid))
        )

        # Iterate over the returned OID-value pairs
        async for errorIndication, errorStatus, errorIndex, varBinds in objects:
            if errorIndication:
                self.ptprint(f"Error: {errorIndication}", out=Out.ERROR)
                break
            elif errorStatus:
                self.ptprint(f"Error: {errorStatus.prettyPrint()} at {errorIndex}", out=Out.ERROR)
                break
            else:
                for oid, value in varBinds:
                    if self.args.oid_format:
                        oid = oid.prettyPrint()  # Convert OID to string
                    value_type = value.__class__.__name__.upper()  # Get the value type
                    value_str = value.prettyPrint()  # Convert value to string

                    # Format the value type and content
                    if value_type == "OCTET STRING":
                        value_output = f'STRING: "{value_str}"'
                    elif value_type == "OBJECT-IDENTIFIER":
                        value_output = f'OID: {value}'
                    elif value_type == "TIMETICKS":
                        value_output = f'Timeticks: ({value_str}) {self.format_timeticks(value)}'
                    elif value_type == "INTEGER":
                        value_output = f'INTEGER: {value_str}'
                    else:
                        value_output = value_str  # Default for other types

                    # Construct the final formatted string
                    formatted_output = f"{oid} = {value_output}"
                    self.ptprint(formatted_output)
                results= "success"

        if self.args.output:
            self.write_to_file(results)
        return results
    
    def output(self) -> None:
        """
        class SNMPResult:
            version: Optional[SNMPVersion] = None
            communities: Optional[List[str]] = None
            usernames: Optional[List[str]] = None
            credentials: Optional[List[Credential]] = None
            Writetest3: Optional[List[WriteTestResult]] = None
            Writetest2: Optional[List[WriteTestResult]] = None
            Bulk2: Optional[List[str]] = None
            Bulk3: Optional[List[str]] = None
        """
        def credentials_to_string(creds: List[Credential]) -> str:
            return ", ".join(
                f"{c.username or 'None'}:{c.password or 'None'}"
                for c in creds
            )
        def write_results_to_string(results: List[WriteTestResult]) -> str:
            return ", ".join(
                f"{str(r.OID) or 'None'}-{r.value or 'None'}-{r.creds}"
                for r in results
            )

        bulk3_str = None
        if self.results.Bulk3 is not None and len(self.results.Bulk3) != 0:
            bulk3_str = self.results.Bulk3 if isinstance(self.results.Bulk3, str) else "\n".join(self.results.Bulk3)
        bulk2_str = None
        if self.results.Bulk2 is not None and len(self.results.Bulk2) != 0:
            bulk2_str = self.results.Bulk2 if isinstance(self.results.Bulk2, str) else "\n".join(self.results.Bulk2)

        properties = {
            "software_type": None,
            "name": "snmp",
            "version": None,
            "vendor": None,
            "description": None,
            "communities": ",".join(self.results.communities) if self.results.communities is not None and len(self.results.communities) != 0 else None,
            "usernames": ",".join(self.results.usernames) if self.results.usernames is not None and len(self.results.usernames) != 0 else None,
            "credentials": credentials_to_string(self.results.credentials) if self.results.credentials is not None and len(self.results.credentials) != 0 else None,
            "writetest3": write_results_to_string(self.results.Writetest3) if self.results.Writetest3 is not None and len(self.results.Writetest3) != 0 else None,
            "writetest2": write_results_to_string(self.results.Writetest2) if self.results.Writetest2 is not None and len(self.results.Writetest2) != 0 else None,
            "bulk3": bulk3_str,
            "bulk2": bulk2_str,
        }
        deferred_vulns = []

        if self.results.communities is not None and len(self.results.communities) != 0:
            deferred_vulns.append({"vuln_code": VULNS.WeakCommunityName.value, "vuln_request": "Bruteforcing SNMPv1-2 community strings", "vuln_response": ",".join(self.results.communities)})
        if self.results.usernames is not None and len(self.results.usernames) != 0:
            deferred_vulns.append({"vuln_code": VULNS.WeakUsername.value, "vuln_request": "Bruteforcing SNMPv3 usernames", "vuln_response": ",".join(self.results.usernames)})
        if self.results.credentials is not None and len(self.results.credentials) != 0:
            deferred_vulns.append({"vuln_code": VULNS.WeakCredentials.value, "vuln_request": "Bruteforcing SNMPv3 credentials", "vuln_response": credentials_to_string(self.results.credentials)})
        if self.results.Writetest3 is not None and len(self.results.Writetest3) != 0:
            deferred_vulns.append({"vuln_code": VULNS.Write_3.value, "vuln_request": "Testing write access trough SNMPv3", "vuln_response": write_results_to_string(self.results.Writetest3)})
        if self.results.Writetest2 is not None and len(self.results.Writetest2) != 0:
            deferred_vulns.append({"vuln_code": VULNS.Write_2.value, "vuln_request": "Testing write access trough SNMPv2", "vuln_response": write_results_to_string(self.results.Writetest2)})
        if bulk3_str is not None:
            deferred_vulns.append({"vuln_code": VULNS.Readmib_3.value, "vuln_request": "Testing reading MIB database trough SNMPv3", "vuln_response": bulk3_str})
        if bulk2_str is not None:
            deferred_vulns.append({"vuln_code": VULNS.Readmib_2.value, "vuln_request": "Testing reading MIB database trough SNMPv3", "vuln_response": bulk2_str})

        snmp_node = self.ptjsonlib.create_node_object("software", None, None, properties)
        self.ptjsonlib.add_node(snmp_node)
        node_key = snmp_node["key"]
        for v in deferred_vulns:
            self.ptjsonlib.add_vulnerability(node_key=node_key, **v)

        self.ptprint(self.ptjsonlib.get_result_json(), json=True)


import setuptools

with open("ptsrvtester/_version.py") as f:
    __version__ = f.readline().split('"')[1]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ptsrvtester",
    description="Application server penetration testing tool (Penterep tool)",
    version=__version__,
    author="Penterep",
    author_email="martsuchwork@gmail.com",
    url="https://www.penterep.com/",
    license="GPLv3+",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Environment :: Console",
        "Topic :: Security",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    python_requires=">=3.12",
    install_requires = [
    "cryptography>=42.0.8",
    "dnspython>=2.7.0",
    "impacket>=0.12.0",
    "ldap3>=2.9.1",
    "ptlibs>=1.0.20,<2",
    "pysnmp>=7.1.20",
    "python-whois>=0.9.5",
    "paramiko>=4.0.0",
    "beautifulsoup4>=4.14.0",
    "httpx>=0.28.0",
    "h2>=4.3.0",
    "lxml>=6.0.0",
    "ssh_audit>=3.3.0",
    "ntlm-auth>=1.5.0",
    "scapy>=2.6.1",
    "dhcppython>=0.1.4",
    "pillow>=12.0.0",
    "python-xlib>=0.33",
    "pynput>=1.8.0",
    "XWindows>=0.0.7",
],
    entry_points={"console_scripts": ["ptsrvtester = ptsrvtester.ptsrvtester:main"]},
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
)
